import habutax.enum as enum
from habutax.form import InputForm, Jurisdiction
from habutax.inputs import *

class Form1099INT(InputForm):
    form_name = "1099-int"
    tax_year = 2023
    description = "Form 1099-INT"
    long_description = "Interest Income"
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        inputs = [
            EnumInput('belongs_to', enum.taxpayer_spouse_or_both, description="To whom does this 1099-INT belong?"),
            StringInput('payer', description="Who paid you this interest?"),
            StringInput('recipient', description="To whom was this interest paid? (you or your spouse's name)"),
            FloatInput('box_1', description="Interest income"),
            FloatInput('box_2', description="Early withdrawal penalty"),
            FloatInput('box_3', description="Interest on U.S. Savings Bonds and Treasury obligations"),
            FloatInput('box_4', description="Federal income tax withheld"),
            FloatInput('box_5', description="Investment expenses"),
            FloatInput('box_6', description="Foreign tax paid"),
            StringInput('box_7', description="Foreign country or US possesion"),
            FloatInput('box_8', description="Tax-exempt interest"),
            FloatInput('box_9', description="Specified private activity bond interest"),
            FloatInput('box_10', description="Market discount"),
            FloatInput('box_11', description="Bond premium"),
            FloatInput('box_12', description="Bond premium on Treasury obligations"),
            FloatInput('box_13', description="Bond premium on tax-exempt bond"),
            StringInput('box_14', description="Tax-exempt and tax credit bond CUSIP no."),
            EnumInput('box_15_1', enum.us_states, allow_empty=True, description="State #1 for which tax was withheld"),
            StringInput('box_16_1', description="State #1 identification no."),
            FloatInput('box_17_1', description="State #1 tax withheld"),
            EnumInput('box_15_2', enum.us_states, allow_empty=True, description="State #2 for which tax was withheld"),
            StringInput('box_16_2', description="State #2 identification no."),
            FloatInput('box_17_2', description="State #2 tax withheld"),
        ]

        super().__init__(__class__, inputs, **kwargs)



import os

from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form8959(Form):
    form_name = "8959"
    tax_year = 2023
    description = "Form 8959"
    long_description = "Additional Medicare Tax"
    jurisdiction = Jurisdiction.US
    sequence_no = 71

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('other_medicare_income', description="Do you have any unreported tip income (Form 4137, line 6) or wages on which medicare tax was not withheld? (this is not common)"),
        ]

        required_fields = [
            # Part I - Additional Medicare Tax on Medicare Wages
            FloatField('1', lambda s, i, v: float(sum([v[f'w-2:{n}.box_5'] for n in range(i['1040.number_w-2'])]))),
            FloatField('2', lambda s, i, v: s.not_implemented() if i['other_medicare_income'] else None),
            FloatField('3', lambda s, i, v: s.not_implemented() if i['other_medicare_income'] else None),
            FloatField('4', lambda s, i, v: v['1'] + v['2'] + v['3']),
            FloatField('5', lambda s, i, v: s.form('1040').threshold('additional_medicare_tax_applies', i['1040.filing_status'])),
            FloatField('6', lambda s, i, v: max(0.0, v['4'] - v['5'])),
            FloatField('7', lambda s, i, v: v['6'] * 0.009),
        ]

        optional_fields = [
            # Part II - Additional Medicare Tax on Self-Employment Income
            FloatField('8', lambda s, i, v: s.not_implmented() if i['1040.self_employment_income'] else None),
            FloatField('9', lambda s, i, v: s.form('1040').threshold('additional_medicare_tax_applies', i['1040.filing_status'])),
            FloatField('10', lambda s, i, v: v['4']),
            FloatField('11', lambda s, i, v: max(0.0, v['9'] - v['10'])),
            FloatField('12', lambda s, i, v: max(0.0, v['8'] - v['11'])),
            FloatField('13', lambda s, i, v: v['12'] * 0.009),

            # Part III - Additional Medicare Tax on  Railroad Retirement Tax Act (RRTA) Compensation
            FloatField('14', lambda s, i, v: s.not_implmented() if i['1040.rrta_compensation'] else None),
            FloatField('15', lambda s, i, v: s.form('1040').threshold('additional_medicare_tax_applies', i['1040.filing_status'])),
            FloatField('16', lambda s, i, v: max(0.0, v['14'] - v['15'])),
            FloatField('17', lambda s, i, v: v['16'] * 0.009),

            # Part IV - Total Additional Medicare Tax
            FloatField('18', lambda s, i, v: v['7'] + v['13'] + v['17']),

            # Part V - Withholding Reconcilation
            FloatField('19', lambda s, i, v: float(sum([v[f'w-2:{n}.box_6'] for n in range(i['1040.number_w-2'])]))),
            FloatField('20', lambda s, i, v: v['1']),
            FloatField('21', lambda s, i, v: v['20'] * 0.0145),
            FloatField('22', lambda s, i, v: max(0.0, v['19'] - v['21'])),
            FloatField('23', lambda s, i, v: s.not_implmented() if i['1040.rrta_compensation'] else None),
            FloatField('24', lambda s, i, v: v['22'] + v['23']),
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_1[0]', '1040.full_names'),
            TextPDFField('topmostSubform[0].Page1[0].f1_2[0]', '1040.you_ssn', max_length=11),
            TextPDFField('topmostSubform[0].Page1[0].f1_3[0]', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_4[0]', '2'),
            TextPDFField('topmostSubform[0].Page1[0].f1_5[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_6[0]', '4'),
            TextPDFField('topmostSubform[0].Page1[0].f1_7[0]', '5'),
            TextPDFField('topmostSubform[0].Page1[0].f1_8[0]', '6'),
            TextPDFField('topmostSubform[0].Page1[0].f1_9[0]', '7'),
            TextPDFField('topmostSubform[0].Page1[0].f1_10[0]', '8'),
            TextPDFField('topmostSubform[0].Page1[0].f1_11[0]', '9'),
            TextPDFField('topmostSubform[0].Page1[0].f1_12[0]', '10'),
            TextPDFField('topmostSubform[0].Page1[0].f1_13[0]', '11'),
            TextPDFField('topmostSubform[0].Page1[0].f1_14[0]', '12'),
            TextPDFField('topmostSubform[0].Page1[0].f1_15[0]', '13'),
            TextPDFField('topmostSubform[0].Page1[0].f1_16[0]', '14'),
            TextPDFField('topmostSubform[0].Page1[0].f1_17[0]', '15'),
            TextPDFField('topmostSubform[0].Page1[0].f1_18[0]', '16'),
            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '17'),
            TextPDFField('topmostSubform[0].Page1[0].f1_20[0]', '18'),
            TextPDFField('topmostSubform[0].Page1[0].f1_21[0]', '19'),
            TextPDFField('topmostSubform[0].Page1[0].f1_22[0]', '20'),
            TextPDFField('topmostSubform[0].Page1[0].f1_23[0]', '21'),
            TextPDFField('topmostSubform[0].Page1[0].f1_24[0]', '22'),
            TextPDFField('topmostSubform[0].Page1[0].f1_25[0]', '23'),
            TextPDFField('topmostSubform[0].Page1[0].f1_26[0]', '24'),
        ]
        pdf_file = os.path.join(os.path.dirname(__file__), 'f8959.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields,
                         pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



import habutax.enum as enum
from habutax.form import InputForm, Jurisdiction
from habutax.inputs import *

class Form1099G(InputForm):
    form_name = "1099-g"
    tax_year = 2023
    description = "Form 1099-G"
    long_description = "Certain Government Payments"
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        inputs = [
            EnumInput('belongs_to', enum.taxpayer_spouse_or_both, description="To whom does this 1099-G belong?"),
            StringInput('payer', description="Who paid you this amount?"),
            StringInput('recipient', description="To whom was this amount paid? (you or your spouse's name)"),
            FloatInput('box_1', description="Unemployment compensation"),
            FloatInput('box_2', description="State income tax refund, credit or offset"),
            IntegerInput('box_3', description="Refund (box 2 amount) is for tax year"),
            FloatInput('box_4', description="Federal income tax withheld"),
            FloatInput('box_5', description="RTAA payments"),
            FloatInput('box_6', description="Taxable grants"),
            FloatInput('box_7', description="Agriculture payments"),
            BooleanInput('box_8', description="Is box 2 is trade or business income?"),
            FloatInput('box_9', description="Market gain"),
            EnumInput('box_10a_1', enum.us_states, allow_empty=True, description="State #1"),
            StringInput('box_10b_1', description="State #1 identification no."),
            FloatInput('box_11_1', description="State #1 tax withheld"),
            EnumInput('box_10a_2', enum.us_states, allow_empty=True, description="State #2"),
            StringInput('box_10b_2', description="State #2 identification no."),
            FloatInput('box_11_2', description="State #2 tax withheld"),
        ]

        super().__init__(__class__, inputs, **kwargs)



import os

import habutax.enum as enum
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class FormNCD400(Form):
    form_name = "nc_d-400"
    tax_year = 2023
    description = "D-400"
    long_description = "N.C. 2023 Individual Income Tax Return"
    jurisdiction = Jurisdiction.NC
    sequence_no = 0

    def __init__(self, **kwargs):
        inputs = [
            StringInput('county', description="Enter the North Carolina county in which you resided on the last day of 2023. Your county of residence may be different than the county of your mailing address."),
            BooleanInput('out_of_country', description="Were you or, if married filing jointly, your spouse were out of the country on the statutory due date of the return?"),
            BooleanInput('nc_residents', description="Were you, and your spouse if filing jointly, residents of N.C. for the entire year of 2023?"),
            BooleanInput('veteran', description="Are you a veteran?"),
            BooleanInput('spouse_veteran', description="Is your spouse a veteran?"),
            BooleanInput('federal_extension', description="Were you granted an automatic extension to file your 2023 federal income tax return, e.g., Form 1040?"),
            IntegerInput('year_spouse_died', description="Which year did your spouse die?"),
            BooleanInput('additions_to_agi', description="Do you have any additions you are required to add to your federal adjusted gross income for your North Carolina return? If unsure, answer \"yes\" and you will be prompted for them individually."),
            BooleanInput('deductions_from_agi', description="Do you have any deductions you are allowed to make from your federal adjusted gross income for your North Carolina return? If unsure, answer \"yes\" and you will be prompted for them individually."),
            BooleanInput('try_itemizing', description="Would you you like to see if itemizing on your NC State tax return saves you money? Even if you answer \"yes\" here, we will only itemize if it helps you."),
            BooleanInput('partial_year_or_nonresident', description="Were you (or your spouse if filing jointly) a part-year resident or a nonresident of NC? You are a “part-year resident” if you moved to N.C. and became a resident during the tax year, or you moved out of N.C. and became a resident of another state during the tax year. You are a “nonresident” if you were not a resident of N.C. at any time during the tax year."),
            BooleanInput('nc_tax_credits', description="Do you want to claim any NC tax credits? These include credits for income tax paid to another state or country, rehabilitating historic structures, business incentive and energy tax credits, or credits carried over from previous years."),
            BooleanInput('no_consumer_use_tax', description="Do you certify that no Consumer Use Tax is due to North Carolina?"),
            BooleanInput('amended_return', description="Is this an amended NC state return? (HabuTax does not support this)"),
            FloatInput('estimated_tax', description="Enter any estimated North Carolina income tax payments for tax year 2023 (including any portion of your 2020 refund that was applied to your 2023 estimated income tax)."),
            FloatInput('paid_with_extension', description="If you previously applied for a N.C. extension (Form D-410), enter the amount of North Carolina income tax paid with the extension."),
            FloatInput('partnership_payments', description="If you are a nonresident partner in a partnership doing business in North Carolina, enter your share of the income tax paid to North Carolina by the manager of the partnership on your distributive share of the partnership income. Include with your return a copy of Form NC K-1 for Form D-403 provided by the partnership to verify the amount claimed (you must do this outside of HabuTax)"),
            FloatInput('s_corp_payments', description="If you are a nonresident shareholder of an S Corporation doing business in North Carolina, enter your share of the income tax paid to North Carolina by the S Corporation on your distributive share of the S Corporation income. Include with your return a copy of Form NC K-1 for Form CD-401S provided by the S Corporation to verify the amount claimed.(you must do this outside of HabuTax)"),
            BooleanInput('late', description="Are you filing this return after the original due date?"),
            FloatInput('interest_on_underpayment', description="It is possible you owe interest because you have underpaid your NC income tax. Please complete NC Form D-422 to determine the amount, if any, of interest you owe and enter that amount here."),
            FloatInput('2024_estimated_income_tax', description="Enter the amount, if any, of your NC state refund you want applied to 2024 estimated tax"), 
            FloatInput('nc_nongame_endangered_wildlife', description="Enter the amount, if any, of your NC state refund you want to contribute to the NC Nongame and Endangered Wildlife Fund"), 
            FloatInput('nc_education_endowment', description="Enter the amount, if any, of your NC state refund you want to contribute to the N.C Education Endowment Fund"), 
            FloatInput('nc_breast_cervical_cancer', description="Enter the amount, if any, of your NC state refund you want to contribute to the N.C Breast and Cervical Cander Control Program"), 
        ]

        def tax_withheld(self, i, v, belongs_to):
            nc_withholding = 0.0
            for n in range(i['1040.number_w-2']):
                if v[f'w-2:{n}.belongs_to'] not in belongs_to:
                    continue
                if v[f'w-2:{n}.box_15'] == enum.us_states.NC:
                    nc_withholding += v[f'w-2:{n}.box_17']
            for n in range(i['1040.number_1099-g']):
                if v[f'1099-g:{n}.belongs_to'] not in belongs_to:
                    continue
                if v[f'1099-g:{n}.box_10a_1'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-g:{n}.box_11_1']
                if v[f'1099-g:{n}.box_10a_2'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-g:{n}.box_11_2']
            for n in range(i['1040.number_1099-int']):
                if v[f'1099-int:{n}.belongs_to'] not in belongs_to:
                    continue
                if v[f'1099-int:{n}.box_15_1'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-int:{n}.box_17_1']
                if v[f'1099-int:{n}.box_15_2'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-int:{n}.box_17_2']
            for n in range(i['1040.number_1099-div']):
                if v[f'1099-div:{n}.belongs_to'] not in belongs_to:
                    continue
                if v[f'1099-div:{n}.box_14_1'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-div:{n}.box_16_1']
                if v[f'1099-div:{n}.box_14_2'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-div:{n}.box_16_2']
            for n in range(i['1040.number_1099-r']):
                if v[f'1099-r:{n}.belongs_to'] not in belongs_to:
                    continue
                if v[f'1099-r:{n}.box_14_1_state'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-r:{n}.box_14_1']
                if v[f'1099-r:{n}.box_14_2_state'] == enum.us_states.NC:
                    nc_withholding += v[f'1099-r:{n}.box_14_2']
            return nc_withholding

        def your_tax_withheld(self, i, v):
            return tax_withheld(self, i, v, [enum.taxpayer_or_spouse.taxpayer, enum.taxpayer_spouse_or_both.taxpayer, enum.taxpayer_spouse_or_both.both])

        def spouses_tax_withheld(self, i, v):
            return tax_withheld(self, i, v, [enum.taxpayer_or_spouse.spouse, enum.taxpayer_spouse_or_both.spouse])

        required_fields = [
            StringField('your_ssn', lambda s, i, v: i['1040.you_ssn'][:3] + "-" + i['1040.you_ssn'][3:5] + "-" + i['1040.you_ssn'][5:]),
            StringField('spouse_ssn', lambda s, i, v: i['1040.spouse_ssn'][:3] + "-" + i['1040.spouse_ssn'][3:5] + "-" + i['1040.spouse_ssn'][5:] if i['1040.filing_status'] == enum.filing_status.MarriedFilingJointly else None),
            StringField('your_first_name', lambda s, i, v: i['1040.first_name'].upper()),
            StringField('your_middle_initial', lambda s, i, v: i['1040.middle_initial'].upper()),
            StringField('your_last_name', lambda s, i, v: i['1040.last_name'].upper()),
            StringField('spouse_first_name', lambda s, i, v: i['1040.spouse_first_name'].upper() if i['1040.filing_status'] == enum.filing_status.MarriedFilingJointly else None),
            StringField('spouse_middle_initial', lambda s, i, v: i['1040.spouse_middle_initial'].upper() if i['1040.filing_status'] == enum.filing_status.MarriedFilingJointly else None),
            StringField('spouse_last_name', lambda s, i, v: i['1040.spouse_last_name'].upper() if i['1040.filing_status'] == enum.filing_status.MarriedFilingJointly else None),
            StringField('mailing_address', lambda s, i, v: i['1040.home_address'].upper()),
            StringField('apartment_no', lambda s, i, v: i['1040.apartment_no'].upper()),
            StringField('city', lambda s, i, v: i['1040.city'].upper()),
            StringField('state', lambda s, i, v: str(i['1040.state'])),
            StringField('zip', lambda s, i, v: i['1040.zip'].upper()),
            StringField('country', lambda s, i, v: i['1040.foreign_country'].upper()),
            StringField('county', lambda s, i, v: i['county'].upper()[:5]),
            StringField('phone_number', lambda s, i, v: i['1040.phone_number']),
            BooleanField('out_of_country', lambda s, i, v: i['out_of_country']),
            BooleanField('nc_residents', lambda s, i, v: i['nc_residents']),
            BooleanField('veteran', lambda s, i, v: i['veteran']),
            BooleanField('spouse_veteran', lambda s, i, v: i['spouse_veteran'] if i['1040.filing_status'] == enum.filing_status.MarriedFilingJointly else None),
            BooleanField('federal_extension', lambda s, i, v: i['federal_extension']),
            BooleanField('1', lambda s, i, v: i['1040.filing_status'] == enum.filing_status.Single),
            BooleanField('2', lambda s, i, v: i['1040.filing_status'] == enum.filing_status.MarriedFilingJointly),
            BooleanField('3', lambda s, i, v: i['1040.filing_status'] == enum.filing_status.MarriedFilingSeparately),
            BooleanField('4', lambda s, i, v: i['1040.filing_status'] == enum.filing_status.HeadOfHousehold),
            BooleanField('5', lambda s, i, v: i['1040.filing_status'] == enum.filing_status.QualifyingSurvivingSpouse),
            StringField('separate_spouse_name', lambda s, i, v: f'i["1040.first_name"] i["1040.middle_initial"] i["1040.last_name"]'.upper() if v['3'] else None),
            StringField('separate_spouse_ssn', lambda s, i, v: i['1040.spouse_ssn'][:3] + "-" + i['1040.spouse_ssn'][3:5] + "-" + i['1040.spouse_ssn'][5:] if v['3'] else None),
            StringField('year_spouse_died', lambda s, i, v: i['year_spouse_died'] if v['5'] else None),
            FloatField('6', lambda s, i, v: v['1040.11'], places=0),
            FloatField('7', lambda s, i, v: v['nc_d-400_ss.15'] if i['additions_to_agi'] else None, places=0),
            FloatField('8', lambda s, i, v: v['6'] + v['7'], places=0),
            FloatField('9', lambda s, i, v: v['nc_d-400_ss.41'] if i['deductions_from_agi'] else None, places=0),
            IntegerField('10a', lambda s, i, v: v['nc_d-400_child_deduction_wkst.3']),
            FloatField('10b', lambda s, i, v: v['nc_d-400_child_deduction_wkst.5'] if v['10a'] > 0 else None, places=0),
            BooleanField('11_itemizing', lambda s, i, v: i['try_itemizing'] and v['nc_d-400_sa.deduction'] > v['nc_d-400_sa.nc_standard_deduction']),
            FloatField('11', lambda s, i, v: v['nc_d-400_sa.deduction'] if v['11_itemizing'] else v['nc_d-400_sa.nc_standard_deduction'], places=0),
            FloatField('12a', lambda s, i, v: v['9'] + v['10b'] + v['11'], places=0),
            FloatField('12b', lambda s, i, v: v['8'] - v['12a'], places=0),
            FloatField('13', lambda s, i, v: s.not_implemented() if i['partial_year_or_nonresident'] else None, places=0),
            FloatField('14', lambda s, i, v: s.not_implemented() if i['partial_year_or_nonresident'] else v['12b'], places=0),
            FloatField('15', lambda s, i, v: v['14'] * 0.0475, places=0), # NC Income Tax
            FloatField('16', lambda s, i, v: s.not_implemented() if i['nc_tax_credits'] else None, places=0),
            FloatField('17', lambda s, i, v: v['15'] - v['16'], places=0),
            BooleanField('no_consumer_use_tax', lambda s, i, v: i['no_consumer_use_tax']),
            FloatField('18', lambda s, i, v: None if v['no_consumer_use_tax'] else v['nc_d-400_consumer_use_tax_wkst.consumer_use_tax'], places=0),
            FloatField('19', lambda s, i, v: v['17'] + v['18'], places=0),
            FloatField('20a', your_tax_withheld, places=0),
            FloatField('20b', spouses_tax_withheld, places=0),
            FloatField('21a', lambda s, i, v: i['estimated_tax'], places=0),
            FloatField('21b', lambda s, i, v: i['paid_with_extension'], places=0),
            FloatField('21c', lambda s, i, v: i['partnership_payments'], places=0),
            FloatField('21d', lambda s, i, v: i['s_corp_payments'], places=0),
            FloatField('22', lambda s, i, v: s.not_implemented() if i['amended_return'] else None, places=0),
            FloatField('23', lambda s, i, v: v['20a'] + v['20b'] + v['21a'] + v['21b'] + v['21c'] + v['21d'] + v['22'], places=0),
            FloatField('24', lambda s, i, v: s.not_implemented() if i['amended_return'] else None, places=0),
            FloatField('25', lambda s, i, v: v['23'] - v['24'], places=0),
            FloatField('refund', lambda s, i, v: v['34'] if v['19'] <= v['25'] else -v['27'], places=0), # Fake "field" to ensure only correct portion is filled
        ]

        optional_fields = [
            # Tax due
            FloatField('26a', lambda s, i, v: v['19'] - v['25'] if v['25'] < v['19'] else s.not_implemented(), places=0),
            FloatField('26b', lambda s, i, v: s.not_implemented() if i['late'] else None, places=0),
            FloatField('26c', lambda s, i, v: s.not_implemented() if i['late'] else None, places=0),
            FloatField('26d', lambda s, i, v: v['26b'] + v['26c'], places=0),
            FloatField('26e', lambda s, i, v: i['interest_on_underpayment'] if (v['17'] - (v['20a'] + v['20b']) >= 1000) else None, places=0),
            FloatField('27', lambda s, i, v: v['26a'] + v['26d'] + v['26e'], places=0),

            # Overpayment
            FloatField('28', lambda s, i, v: v['25'] - v['19'] if v['25'] >= v['19'] else s.not_implemented(), places=0),
            FloatField('29', lambda s, i, v: i['2024_estimated_income_tax'], places=0),
            FloatField('30', lambda s, i, v: i['nc_nongame_endangered_wildlife'], places=0),
            FloatField('31', lambda s, i, v: i['nc_education_endowment'], places=0),
            FloatField('32', lambda s, i, v: i['nc_breast_cervical_cancer'], places=0),
            FloatField('33', lambda s, i, v: v['29'] + v['30'] + v['31'] + v['32'], places=0),
            FloatField('34', lambda s, i, v: v['28'] - v['33'] if v['28'] >= v['33'] else s.not_implemented(), places=0),
        ]

        pdf_fields = [
#            TextPDFField('PRINT', 'OTHER THAN YOUR SIGNATURE, DO NOT HANDWRITE ON THIS FORM'),
            #OptionlessButtonPDFField('z_d400wf_clear', 'unknown'),
            #OptionlessButtonPDFField('z_d400wf_print', 'unknown'),
#            TextPDFField('y_d400wf_datebeg', 'fiscal_year_beginning'),
#            ButtonPDFField('y_d400wf_amendedreturn', 'unknown', 'Yes'),
#            TextPDFField('y_d400wf_dateend', 'fiscal_year_ending'),
            TextPDFField('y_d400wf_ssn1', 'your_ssn'),
            TextPDFField('y_d400wf_ssn2', 'spouse_ssn'),
            TextPDFField('y_d400wf_fname1', 'your_first_name', max_length=25),
            TextPDFField('y_d400wf_mi1', 'your_middle_initial', max_length=1),
            TextPDFField('y_d400wf_lname1', 'your_last_name', max_length=25),
            TextPDFField('y_d400wf_fname2', 'spouse_first_name', max_length=25),
            TextPDFField('y_d400wf_mi2', 'spouse_middle_initial', max_length=1),
            TextPDFField('y_d400wf_lname2', 'spouse_last_name', max_length=25),
            TextPDFField('y_d400wf_add', 'mailing_address', max_length=50),
            TextPDFField('y_d400wf_apartment number', 'apartment_no', max_length=5),
            TextPDFField('y_d400wf_city', 'city', max_length=27),
            ChoicePDFField('y_d400wf_state', 'state', [' ', 'AK', 'AL', 'AR', 'AZ', 'CA', 'CO', 'CT', 'DC', 'DE', 'FL', 'GA', 'HI', 'IA', 'ID', 'IL', 'IN', 'KS', 'KY', 'LA', 'MA', 'MD', 'ME', 'MI', 'MN', 'MO', 'MS', 'MT', 'NC', 'ND', 'NE', 'NH', 'NJ', 'NM', 'NV', 'NY', 'OH', 'OK', 'OR', 'PA', 'PR', 'RI', 'SC', 'SD', 'TN', 'TX', 'UT', 'VA', 'VT', 'WA', 'WI', 'WV', 'WY']),
            TextPDFField('y_d400wf_zip', 'zip', max_length=10),
            TextPDFField('y_d400wf_country', 'country', max_length=9),
            TextPDFField('y_d400wf_county', 'county', max_length=5),
#            TextPDFField('y_d400wf_NCEDU_good', 'nc_education_endowment_fund', max_length=9),
            ButtonPDFField('y_d400wf_Out of Country', 'out_of_country', 'Yes'),
#            ButtonPDFField('y_d400wf_Signed by Executor or Administrator', 'unknown', 'Yes'),
#            TextPDFField('y_d400wf_dead1', 'unknown'),
#            TextPDFField('y_d400wf_dead2', 'unknown'),
            ButtonPDFField('y_d400wf_rs1yes', 'nc_residents', 'Yes'),
            ButtonPDFField('y_d400wf_rs1no', 'nc_residents', 'Yes', lambda s, v, f: not v),
            ButtonPDFField('y_d400wf_rs2yes', 'nc_residents', 'Yes'),
            ButtonPDFField('y_d400wf_rs2no', 'nc_residents', 'Yes', lambda s, v, f: not v),
            ButtonPDFField('y_d400wf_v1yes', 'veteran', 'Yes'),
            ButtonPDFField('y_d400wf_v2no', 'veteran', 'Yes', lambda s, v, f: not v),
            ButtonPDFField('y_d400wf_sv1yes', 'spouse_veteran', 'Yes'),
            ButtonPDFField('y_d400wf_sv1no', 'spouse_veteran', 'Yes', lambda s, v, f: not v),
            ButtonPDFField('y_d400wf_fedex1yes', 'federal_extension', 'Yes'),
            ButtonPDFField('y_d400wf_fedex1no', 'federal_extension', 'Yes', lambda s, v, f: not v),
            ButtonPDFField('y_d400wf_fstat1', '1', 'Yes'),
            ButtonPDFField('y_d400wf_fstat2', '2', 'Yes'),
            ButtonPDFField('y_d400wf_fstat3', '3', 'Yes'),
            ButtonPDFField('y_d400wf_fstat4', '4', 'Yes'),
            ButtonPDFField('y_d400wf_fstat5', '5', 'Yes'),
            TextPDFField('y_d400wf_sname2', 'separate_spouse_name', max_length=36),
            TextPDFField('y_d400wf_sssn2', 'separate_spouse_ssn'),
#            ChoicePDFField('y_d400wf_dead3', 'year_spouse_died', [' ', '2021', '2022']),
            TextPDFField('y_d400wf_li6_good', '6', max_length=9),
            TextPDFField('y_d400wf_li7_good', '7', max_length=8),
            TextPDFField('y_d400wf_li8_good', '8', max_length=9),
            TextPDFField('y_d400wf_li9_good', '9', max_length=8),
            TextPDFField('y_d400wf_li10a_good', '10a', max_length=2),
            TextPDFField('y_d400wf_li10b_good', '10b', max_length=5),
            ButtonPDFField('y_d400wf_ncstandarddeduction', '11_itemizing', 'Yes', lambda s, v, f: not v),
            ButtonPDFField('y_d400wf_ncitemizeddeduction', '11_itemizing', 'Yes'),
            TextPDFField('y_d400wf_li11_page1_good', '11', max_length=8),
            TextPDFField('y_d400wf_li12a_pg1_good', '12a', max_length=9),
            TextPDFField('y_d400wf_li12b_pg1_good', '12b', max_length=9),
            TextPDFField('y_d400wf_li13_page1_good', '13', max_length=6),
            TextPDFField('y_d400wf_li14_pg1_good', '14', max_length=9),
            TextPDFField('y_d400wf_li15_pg1_good', '15', max_length=8),
            TextPDFField('y_d400wf_lname2_PG2', 'your_last_name', max_length=10),
            TextPDFField('y_d400wf_li16_pg2_good', '16', max_length=8),
            TextPDFField('y_d400wf_li17_pg2_good', '17', max_length=8),
            ButtonPDFField('y_d400wf_Consumer Use Tax', 'no_consumer_use_tax', 'Yes'),
            TextPDFField('y_d400wf_li18_pg2_good', '18', max_length=8),
            TextPDFField('y_d400wf_li19_pg2_good', '19', max_length=8),
            TextPDFField('y_d400wf_li20a_pg2_good', '20a', max_length=8),
            TextPDFField('y_d400wf_li20b_pg2_good', '20b', max_length=8),
            TextPDFField('y_d400wf_li21a_pg2_good', '21a', max_length=8),
            TextPDFField('y_d400wf_li21b_pg2_good', '21b', max_length=8),
            TextPDFField('y_d400wf_li21c_pg2_good', '21c', max_length=8),
            TextPDFField('y_d400wf_li21d_pg2_good', '21d', max_length=8),
            TextPDFField('y_d400wf_li22_pg2_good', '22', max_length=8),
            TextPDFField('y_d400wf_li23_pg2_good', '23', max_length=8),
            TextPDFField('y_d400wf_li24_pg2_good', '24', max_length=8),
            TextPDFField('y_d400wf_li25_pg2_good', '25', max_length=9),
            TextPDFField('y_d400wf_li26a_pg2_good', '26a', max_length=8),
            TextPDFField('y_d400wf_li26b_pg2_good', '26b', max_length=6),
            TextPDFField('y_d400wf_li26c_pg2_good', '26c', max_length=6),
            TextPDFField('y_d400wf_li26d_pg2_good', '26d', max_length=8),
            TextPDFField('y_d400wf_li26e_pg2_good', '26e', max_length=8),
            TextPDFField('y_d400wf_li27_pg2_good', '27', max_length=8),
#            ChoicePDFField('y_d400wf_Ex_good', 'unknown', [' ', 'A', 'F']),
            TextPDFField('y_d400wf_li28_pg2_good', '28', max_length=8),
            TextPDFField('y_d400wf_li29_pg2_good', '29', max_length=8),
            TextPDFField('y_d400wf_li30_pg2_good', '30', max_length=8),
            TextPDFField('y_d400wf_li31_pg2_good', '31', max_length=8),
            TextPDFField('y_d400wf_li32_pg2_good', '32', max_length=8),
            TextPDFField('y_d400wf_li33_pg2_good', '33', max_length=8),
            TextPDFField('y_d400wf_li34_pg2_good', '34', max_length=8),
#            TextPDFField('y_d400wf_sigdate', 'date'),
#            TextPDFField('y_d400wf_sigdate2', 'spouse_date'),
            TextPDFField('y_d400wf_dayphone', 'phone_number'),
#            ButtonPDFField('y_d400wf_check', 'prep_authorization', 'Yes'),
#            TextPDFField('y_d400wf_ppsigdate', 'prep_date'),
#            TextPDFField('y_d400wf_ppssn', 'prep_id', max_length=9),
#            TextPDFField('y_d400wf_ppphoneno', 'prep_phone'),
        ]
        pdf_file = os.path.join(os.path.dirname(__file__), 'fnc_d-400.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



from habutax.enum import filing_status as status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *

class Form1040S2Need6251(Form):
    form_name = "1040_s2_need_6251"
    tax_year = 2023
    description = "Worksheet To See if You Should Fill in Form 6251 (Form 1040)"
    long_description = "Schedule 2, Line 1"
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        thresholds = {
            'line_6': {
                (status.Single, status.HeadOfHousehold):                          81300.0,
                (status.MarriedFilingJointly, status.QualifyingSurvivingSpouse): 126500.0,
                status.MarriedFilingSeparately:                                   63250.0,
            },
            'line_8': {
                (status.Single, status.HeadOfHousehold):                          578150.0,
                (status.MarriedFilingJointly, status.QualifyingSurvivingSpouse): 1156300.0,
                status.MarriedFilingSeparately:                                   578150.0,
            },
            'line_12_comparison': {
                (status.Single, status.HeadOfHousehold,
                 status.MarriedFilingJointly,
                 status.QualifyingSurvivingSpouse): 220700.0,
                status.MarriedFilingSeparately:     110350.0,
            }
        }

        inputs = [
            BooleanInput('schedule_j_needed', description="Do you need to fill our Schedule J? (Income Averaging for Farmers and Fishermen)"),
            BooleanInput('form_4972_needed', description="Do you need to fill out Form 4972? (Tax on Lump-Sum Distributions From Qualified Plans of Participants Born Before January 2, 1936)"),
        ]

        def line_13(self, i, v):
            if i['schedule_j_needed'] or i['form_4972_needed']:
                self.not_implemented()
            amount = v['1040.16']
            if i['1040.need_8962']:
                amount += v['1040_s2.2']
            return amount

        def need_6251(self, i, v):
            if not v['5'] > v['6']:
                return False
            elif v['11'] > self.threshold('line_12_comparison', i['1040.filing_status']):
                return True
            return v['12'] > v['13']

        optional_fields = [
            FloatField('1', lambda s, i, v: s.not_implemented() if not v['1040.itemizing'] else v['1040.15']),
            FloatField('2', lambda s, i, v: s.not_implemented() if not v['1040.itemizing'] else v['1040_sa.7']),
            FloatField('3', lambda s, i, v: v['1040.11'] - v['1040.13'] if not v['1040.itemizing'] else v['1'] + v['2']),
            FloatField('4', lambda s, i, v: v['1040_s1.1'] + v['1040_s1.8z'] if v['1040.schedule_1_additional_income'] else None),
            FloatField('5', lambda s, i, v: v['3'] - v['4']),
            FloatField('6', lambda s, i, v: s.threshold('line_6', i['1040.filing_status'])),
            FloatField('7', lambda s, i, v: v['5'] - v['6']),
            FloatField('8', lambda s, i, v: s.threshold('line_8', i['1040.filing_status'])),
            FloatField('9', lambda s, i, v: 0.0 if v['5'] > v['8'] else v['5'] - v['8']),
            FloatField('10', lambda s, i, v: min(v['6'], v['9'] * 0.25)),
            FloatField('11', lambda s, i, v: v['7'] + v['10'] if v['5'] > v['8'] else v['7']),
            FloatField('12', lambda s, i, v: v['11'] * 0.26),
            FloatField('13', line_13),
            BooleanField('need_6251', need_6251),
        ]

        super().__init__(__class__, inputs, [], optional_fields,
                         thresholds=thresholds, **kwargs)

    def needs_filing(self, values):
        return False



import os

from habutax.enum import filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form1040S1(Form):
    form_name = "1040_s1"
    tax_year = 2022
    description = "Schedule 1 (Form 1040)"
    long_description = "Additional Income and Adjustments to Income"
    jurisdiction = Jurisdiction.US
    sequence_no = 1

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('state_local_income_tax_adjust', description="Do you have any refunds, credits, or offsets of state or local income taxes you received in 2022 which you are required to report which you have not already reported via entered 1099-G forms, or adjustments to make to the amounts on any forms 1099-G you have or will enter? For example, answer 'yes' if you did not itemize on your federal tax return in the year to which a refund pertains."),
            FloatInput('state_local_income_tax', description="Enter the total amount of any refunds, credits, or offsets of state or local income taxes you received in 2022 which you are required to report (see instructions). This typically includes amounts you received on form 1099-G unless you did not itemize in the year the refund is for. Please include the amounts even from Forms 1099-G you have entered or will enter into HabuTax."),
            FloatInput('alimony_received', description="Enter amounts received as alimony or separate maintenance pursuant to a divorce or separation agreement entered into on or before December 31, 2018, unless that agreement was changed after December 31, 2018, to expressly provide that alimony received isn't included in your income"),
            StringInput('alimony_received_date', description="Date of original divorce or separation agreement for alimony received."),
            BooleanInput('business_income', description="Did you (or your spouse if filing a joint return) operate a business or practice your profession as a sole proprietor?"),
            BooleanInput('other_gains_losses', description="Do you have gains or losses from selling or exchanging assets used in a trade or business in 2022?"),
            BooleanInput('real_estate', description="Do you have income to report from rental real estate, royalties, partnerships, S corporations, trusts, etc.?"),
            BooleanInput('farm_income_loss', description="Do you have farm income (or loss) to report for 2022?"),
            FloatInput('unemployment_income', description="Enter the amount of any unemployment income you received in 2022 (likely reported on a 1099-G), reduced by any amount you repaid or contributions made to a governmental unemployment compensation program (see instructions for Schedule 1, line 7). Note that any unemployment income entered into HabuTax via a 1099-G is not being considered unless you enter it again here."),
            BooleanInput('uncommon_income', description="Do you have less common income to report for 2022? This might include net operating losses, gambling, cancellation of debt, foreign earned income, taxable HSA distributions, Alaska Permanent Fund dividends, jury duty pay, prizes and awards, activity not engaged in for profit income, stock options, income from the rental of personal property if you engaged in the rental for profit but were not in the business of renting such property, Olympic/Paralympic medals and USOC prize money, section 951(c) inclusion, section 951A(a) inclusion, Section 461(l) excess business loss adjustment, taxable distributions from an ABLE account, scholarship and fellowship grants not reported on Form W-2, Medicaid waiver payments, pension or annuity from a nonqualifed deferred compensation plan or a nongovernmental section 457 plan, or wages earned while incarcerated? (see instructions for Form 1040 Schedule 1, lines 8a-8u, 8z)"),
            BooleanInput('need_other_income', description="Do you have other income to report (not covered by \"less common income\")? Do not include refunds of overpaid interest reported on form 1098, box 4 (unless you are not entering that form into HabuTax)"),
            StringInput('other_income_type', description="Description of any other income you need to report"),
            FloatInput('other_income_amount', description="Amount of other income you need to report"),
            FloatInput('educator_expenses', description=f'If you or your spouse were an eligible educator in {Form1040S1.tax_year}, enter up to $250 per educator in qualified educator expenses. Before entering, reduce this amount if you had any excludable US series EE and I savings bond intereset, nontaxable qualified tuition program earnings/distributions, nontaxable distribution of Coverdel education savings account earnings, or were reimbursed for these expenses in a way not reported on box 1 of your Form W-2.'),
            BooleanInput('certain_business_expenses', description='Did you incurr any of the following expenses in 2022: Certain business expenses of National Guard and reserve members who traveled more than 100 miles from home to perform services as a National Guard or reserve member; Performing-arts-related expenses as a qualified performing artist; Business expenses of fee-basis state or local government officials. For more details, see Form 2106.'),
            BooleanInput('moving_expenses', description='Did you have any moving expenses as a member of the Armed Forces on active duty and due to a military order you move because of a permanent change of station in 2022? See form 3903 and Tax Topic 455.'),
            BooleanInput('deductible_self_employment_tax', description='Were you self-employed and did you self-employment tax in 2022? See Form 1040, Schedule SE.'),
            BooleanInput('sep_simple_qualified', description="Were you self-employed and eligible to deduct any contributions to SEP, SIMPLE, or Qualified Plans in 2022?"),
            BooleanInput('self_employed_health_insurance', description="Are you self-employed and able to deduct the amount you paid for health insurance? (See Form 1040, Schedule 1, line 17 instructions."),
            FloatInput('alimony_paid', description="Enter amounts paid as alimony or separate maintenance pursuant to a divorce or separation agreement entered into on or before December 31, 2018, unless that agreement was changed after December 31, 2018, to expressly provide that alimony paid isn't included in your former spouse's income"),
            SSNInput('alimony_paid_ssn', description="SSN of recipient of alimony paid."),
            StringInput('alimony_paid_date', description="Date of original divorce or separation agreement for alimony paid."),
            FloatInput('traditional_ira_deduction', description="If you (and/or your spouse if filing jointly) are eligible to take a deduction for contributions to a traditional IRA made in 2022, enter that amount here. See instructions for Form 1040, Schedule 1, line 20 for more information"),
            BooleanInput('student_loan_interest', description="Are you (or your spouse if filing jointly) eligible to take a deduction for student loan interest? See instructions for Form 1040, Schedule 1, line 21 for more details."),
            BooleanInput('archer_msa_deduction', description="Are you (or your spouse if filing jointly) eligible to take a deduction for contributions to an Archer MSA made for 2022? See instructions for Form 1040, Schedule 1, line 23 for more details."),
            BooleanInput('uncommon_deductions', description="Do you have less common decuctions to report for 2022? This might include jury duty pay, deductible expenses related to income reported from the rental of personal property engaged in for profit, nontaxable amount of the value of Olympic and Paralympic medals and USOC prize money, reforestation amortization and expenses, repayment of supplemental unemployment benefits under the Trade Act of 1974, contributions to section 501(c)(18)(D) pension plan, contributions by certain chaplains to section 403(b) plans, attorney fees and court costs for actions involving certain unlawful discrimination claims, attorney fees and court costs you paid in connection with an award from the IRS for information you provided that helped the IRS detect tax law violations, housing deduction from Form 2555, or excess deductions of section 67(e) expenses from Schedule K-1 (Form 1041). (see instructions for Form 1040 Schedule 1, lines 24a-24k, 24z)"),
            BooleanInput('need_other_adjustments', description="Do you have other adjustments to income to report (not covered by \"less common deductions\")?"),
            StringInput('other_adjustments_type', description="Description of any other adjustments to income."),
            FloatInput('other_adjustments_amount', description="Amount of other adjustments to income you need to report"),
            BooleanInput('hsa_contribution_you', description='Did you make a contribution to an HSA and need to file Form 8889?'),
            BooleanInput('hsa_contribution_spouse', description='Did your spouse make a contribution to an HSA and need to file Form 8889?'),
        ]

        def line_1(self, i, v):
            if i['state_local_income_tax_adjust']:
                return i['state_local_income_tax']
            return float(sum([v[f'1099-g:{n}.box_2'] for n in range(i['1040.number_1099-g'])]))

        def other_income(self, i, v):
            mort_int_refund = sum([v[f'1098:{n}.box_4'] for n in range(i['1040.number_1098'])])
            types = []
            income = []
            if mort_int_refund > 0.001:
                types.append("Refund of overpaid mortgage interest")
                income.append(mort_int_refund)
            if i['need_other_income']:
                types.append(i['other_income_type'])
                income.append(i['other_income_amount'])
            if len(types) > 0:
                return (", ".join(types), sum(income))
            return (None, None)

        def line_18(self, i, v):
            penalties = float(sum([v[f'1099-int:{n}.box_2'] for n in range(i['1040.number_1099-int'])]))
            penalties += float(sum([v[f'1099-oid:{n}.box_3'] for n in range(i['1040.number_1099-oid'])]))
            if penalties > 0.01:
                return penalties
            return ""

        def line_13(self, i, v):
            spouse_hsa = (i['1040.filing_status'] == filing_status.MarriedFilingJointly) and i['hsa_contribution_spouse']
            if not i['hsa_contribution_you'] and not spouse_hsa:
                return None

            hsa_deduction = v['8889:you.hsa_deduction'] if i['hsa_contribution_you'] else 0.0
            hsa_deduction += v['8889:spouse.hsa_deduction'] if spouse_hsa else 0.0
            return hsa_deduction

        optional_fields = [
            FloatField('1', line_1),
            FloatField('2a', lambda s, i, v: i['alimony_received']),
            FloatField('3', lambda s, i, v: s.not_implemented() if i['business_income'] else None), # Schedule C
            FloatField('4', lambda s, i, v: s.not_implemented() if i['other_gains_losses'] else None), # Form 4797
            FloatField('5', lambda s, i, v: s.not_implemented() if i['real_estate'] else None), # Schedule E
            FloatField('6', lambda s, i, v: s.not_implemented() if i['farm_income_loss'] else None), # Schedule F
            FloatField('7', lambda s, i, v: i['unemployment_income']),
            FloatField('8a', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8b', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8c', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8d', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8e', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8f', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8g', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8h', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8i', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8j', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8k', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8l', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8m', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8n', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8o', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8p', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8q', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8r', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8s', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8t', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8u', lambda s, i, v: s.not_implemented() if i['uncommon_income'] else None),
            FloatField('8z', lambda s, i, v: other_income(s, i, v)[1]),
            FloatField('9', lambda s, i, v: sum([v[f'8{l}'] for l in "abcdefghijklmnopqrstuz"])),
            FloatField('10', lambda s, i, v: v['1'] + v['2a'] + sum([v[f'{n}'] for n in range(3,8)]) + v['9']),
            FloatField('11', lambda s, i, v: s.not_implemented() if i['educator_expenses'] > 500.00 else i['educator_expenses']),
            FloatField('12', lambda s, i, v: s.not_implemented() if i['certain_business_expenses'] else None),
            FloatField('13', line_13), # Form 8889
            FloatField('14', lambda s, i, v: s.not_implemented() if i['moving_expenses'] else None),
            FloatField('15', lambda s, i, v: s.not_implemented() if i['deductible_self_employment_tax'] else None),
            FloatField('16', lambda s, i, v: s.not_implemented() if i['sep_simple_qualified'] else None),
            FloatField('17', lambda s, i, v: s.not_implemented() if i['self_employed_health_insurance'] else None),
            FloatField('18', line_18),
            FloatField('19a', lambda s, i, v: i['alimony_paid']),
            FloatField('20', lambda s, i, v: i['traditional_ira_deduction']),
            FloatField('21', lambda s, i, v: s.not_implemented() if i['student_loan_interest'] else None),
            FloatField('22', lambda s, i, v: ""),
            FloatField('23', lambda s, i, v: s.not_implemented() if i['archer_msa_deduction'] else None),
            FloatField('24a', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24b', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24c', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24d', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24e', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24f', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24g', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24h', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24i', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24j', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24k', lambda s, i, v: s.not_implemented() if i['uncommon_deductions'] else None),
            FloatField('24z', lambda s, i, v: i['other_adjustments_amount'] if i['need_other_adjustments'] else None),
            FloatField('25', lambda s, i, v: sum([v[f'24{l}'] for l in "abcdefghijkz"])),
            FloatField('26', lambda s, i, v: sum([v[f'{n}'] for n in list(range(11,19)) + list(range(20,24))]) + v['19a'] + v['25']),
        ]
        required_fields = [
            StringField('2b', lambda s, i, v: i['alimony_received_date'] if v['2a'] > 0.001 else None),
            StringField('8z_type', lambda s, i, v: other_income(s, i, v)[0]),
            StringField('19b', lambda s, i, v: i['alimony_paid_ssn'] if v['19a'] > 0.001 else None),
            StringField('19c', lambda s, i, v: i['alimony_paid_date'] if v['19a'] > 0.001 else None),
            StringField('24z_type', lambda s, i, v: i['other_adjustments_type'] if i['need_other_adjustments'] else None),
        ]

        pdf_fields = [
            TextPDFField('form1[0].Page1[0].f1_01[0]', '1040.full_names'),
            TextPDFField('form1[0].Page1[0].f1_02[0]', '1040.you_ssn', max_length=11),
            TextPDFField('form1[0].Page1[0].f1_03[0]', '1'),
            TextPDFField('form1[0].Page1[0].f1_04[0]', '2a'),
            TextPDFField('form1[0].Page1[0].f1_05[0]', '2b'),
            TextPDFField('form1[0].Page1[0].f1_06[0]', '3'),
            TextPDFField('form1[0].Page1[0].f1_07[0]', '4'),
            TextPDFField('form1[0].Page1[0].f1_08[0]', '5'),
            TextPDFField('form1[0].Page1[0].f1_09[0]', '6'),
            TextPDFField('form1[0].Page1[0].f1_10[0]', '7'),
            TextPDFField('form1[0].Page1[0].Line8a_ReadOrder[0].f1_11[0]', '8a'),
            TextPDFField('form1[0].Page1[0].f1_12[0]', '8b'),
            TextPDFField('form1[0].Page1[0].f1_13[0]', '8c'),
            TextPDFField('form1[0].Page1[0].f1_14[0]', '8d'),
            TextPDFField('form1[0].Page1[0].f1_15[0]', '8e'),
            TextPDFField('form1[0].Page1[0].f1_16[0]', '8f'),
            TextPDFField('form1[0].Page1[0].f1_17[0]', '8g'),
            TextPDFField('form1[0].Page1[0].f1_18[0]', '8h'),
            TextPDFField('form1[0].Page1[0].f1_19[0]', '8i'),
            TextPDFField('form1[0].Page1[0].f1_20[0]', '8j'),
            TextPDFField('form1[0].Page1[0].f1_21[0]', '8k'),
            TextPDFField('form1[0].Page1[0].f1_22[0]', '8l'),
            TextPDFField('form1[0].Page1[0].f1_23[0]', '8m'),
            TextPDFField('form1[0].Page1[0].f1_24[0]', '8n'),
            TextPDFField('form1[0].Page1[0].f1_25[0]', '8o'),
            TextPDFField('form1[0].Page1[0].f1_26[0]', '8p'),
            TextPDFField('form1[0].Page1[0].f1_27[0]', '8q'),
            TextPDFField('form1[0].Page1[0].f1_28[0]', '8r'),
            TextPDFField('form1[0].Page1[0].f1_29[0]', '8s'),
            TextPDFField('form1[0].Page1[0].f1_30[0]', '8t'),
            TextPDFField('form1[0].Page1[0].f1_31[0]', '8u'),
#            TextPDFField('form1[0].Page1[0].Line8z_ReadOrder[0].f1_32[0]', '8z_type'),
            TextPDFField('form1[0].Page1[0].Line8z_ReadOrder[0].f1_33[0]', '8z_type'),
            TextPDFField('form1[0].Page1[0].f1_34[0]', '8z'),
            TextPDFField('form1[0].Page1[0].f1_35[0]', '9'),
            TextPDFField('form1[0].Page1[0].f1_36[0]', '10'),
            TextPDFField('form1[0].Page2[0].f2_01[0]', '11'),
            TextPDFField('form1[0].Page2[0].f2_02[0]', '12'),
            TextPDFField('form1[0].Page2[0].f2_03[0]', '13'),
            TextPDFField('form1[0].Page2[0].f2_04[0]', '14'),
            TextPDFField('form1[0].Page2[0].f2_05[0]', '15'),
            TextPDFField('form1[0].Page2[0].f2_06[0]', '16'),
            TextPDFField('form1[0].Page2[0].f2_07[0]', '17'),
            TextPDFField('form1[0].Page2[0].f2_08[0]', '18'),
            TextPDFField('form1[0].Page2[0].f2_09[0]', '19a'),
            TextPDFField('form1[0].Page2[0].Line19b_CombField[0].f2_10[0]', '19b', max_length=9),
            TextPDFField('form1[0].Page2[0].f2_11[0]', '19c'),
            TextPDFField('form1[0].Page2[0].f2_12[0]', '20'),
            TextPDFField('form1[0].Page2[0].f2_13[0]', '21'),
            TextPDFField('form1[0].Page2[0].f2_14[0]', '22'), #"Reserved for future use"
            TextPDFField('form1[0].Page2[0].f2_15[0]', '23'),
            TextPDFField('form1[0].Page2[0].Line24a_ReadOrder[0].f2_16[0]', '24a'),
            TextPDFField('form1[0].Page2[0].f2_17[0]', '24b'),
            TextPDFField('form1[0].Page2[0].f2_18[0]', '24c'),
            TextPDFField('form1[0].Page2[0].f2_19[0]', '24d'),
            TextPDFField('form1[0].Page2[0].f2_20[0]', '24e'),
            TextPDFField('form1[0].Page2[0].f2_21[0]', '24f'),
            TextPDFField('form1[0].Page2[0].f2_22[0]', '24g'),
            TextPDFField('form1[0].Page2[0].f2_23[0]', '24h'),
            TextPDFField('form1[0].Page2[0].f2_24[0]', '24i'),
            TextPDFField('form1[0].Page2[0].f2_25[0]', '24j'),
            TextPDFField('form1[0].Page2[0].f2_26[0]', '24k'),
#            TextPDFField('form1[0].Page2[0].Line24z_ReadOrder[0].f2_27[0]', '24z_type'),
            TextPDFField('form1[0].Page2[0].Line24z_ReadOrder[0].f2_28[0]', '24z_type'),
            TextPDFField('form1[0].Page2[0].f2_29[0]', '24z'),
            TextPDFField('form1[0].Page2[0].f2_30[0]', '25'),
            TextPDFField('form1[0].Page2[0].f2_31[0]', '26'),
        ]
        pdf_file = os.path.join(os.path.dirname(__file__), 'f1040s1.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



from types import MethodType

class FieldNotImplemented(Exception):
    def __init__(self, field_name, message_fmt="Encountered unimplemented tax scenario when processing {field_name}", detailed=None):
        self.field_name = field_name
        self.message = message_fmt.format(field_name=field_name)
        if detailed is not None:
            self.message += f': {detailed}'
        super().__init__(self.message)

class Field(object):
    def __init__(self, name):
        assert "." not in name
        self._name = name

    def __form_init__(self, form):
        """Called when this ConfigInput is associated with a form instance.
        This must be called before any other methods are called on the
        Field."""
        self._form = form

    def form(self, form_name=None):
        if form_name is None:
            return self._form
        else:
            return self._form.solver().forms[form_name]

    def threshold(self, name, requested_key=None):
        return self.form().threshold(name, requested_key=requested_key)

    def base_name(self):
        """Return the name relative to the form it is in"""
        return self._name

    def name(self):
        return f'{self._form.name()}.{self._name}'

    def not_implemented(self, detailed=None):
        """Can be called by a field it encounters a scenario which it does not
        have the logic to return a value for"""
        raise FieldNotImplemented(self.name(), detailed=detailed)

    def value(self, inputs, values):
        raise NotImplementedError()

    def to_string(self, value):
        raise NotImplementedError()

    def from_string(self, string):
        raise NotImplementedError()

class TypedField(Field):
    def __init__(self, name, value_fn, _type):
        self._value = MethodType(value_fn, self)
        self._type = _type
        super().__init__(name)

    def value(self, inputs, values):
        v = self._value(inputs, values)
        if v is None or isinstance(v, str) and v.strip() == "":
            return self._empty_value
        elif type(v) is not self._type:
            raise TypeError(f'Field named {self.name()} expected to produce type {self._type}, but found {type(v)}.')
        return v

class BasicTypedField(TypedField):
    def to_string(self, value):
        return str(value)

    def from_string(self, string):
        return self._type(string)

class StringField(BasicTypedField):
    def __init__(self, name, value_fn):
        self._empty_value = ""
        super().__init__(name, value_fn, str)

class BooleanField(BasicTypedField):
    def __init__(self, name, value_fn):
        self._empty_value = False
        super().__init__(name, value_fn, bool)

    def from_string(self, string):
        return string.strip().lower() == 'true'

class IntegerField(BasicTypedField):
    def __init__(self, name, value_fn):
        self._empty_value = 0
        super().__init__(name, value_fn, int)

class FloatField(BasicTypedField):
    def __init__(self, name, value_fn, places=2):
        self._empty_value = 0.0
        self._places = places
        super().__init__(name, value_fn, float)

    def value(self, inputs, values):
        value = super().value(inputs, values)
        return round(value, self._places)

    def to_string(self, value):
        return f'{value:.{self._places}f}'

    def from_string(self, string):
        return round(float(string), self._places)

class EnumField(TypedField):
    def __init__(self, name, enum, value_fn):
        self._empty_value = None
        super().__init__(name, value_fn, enum)

    def enum(self):
        return self._type

    def to_string(self, value):
        if value is None:
            return ""
        return str(value)

    def from_string(self, string):
        if len(string) == 0:
            return None
        return self.enum()[string]



import configparser
import unittest

import habutax.enum as enum
from habutax.fields import *
from habutax.form import Form


class FieldTestForm(Form):
    form_name = "0000"
    tax_year = 1970

    def __init__(self, **kwargs):
        super().__init__(__class__, [], [], [], **kwargs)


class FieldTestCase(unittest.TestCase):
    def setUp(self):
        self.config = configparser.ConfigParser()
        self.form = FieldTestForm()

    def tearDown(self):
        del self.config
        del self.form

    def test_not_implemented(self):
        f = IntegerField('favorite_number', lambda s, i, v: s.not_implemented() if i['favorite_number'] < 10 else i['favorite_number'])
        f.__form_init__(self.form)

        self.assertEqual(f.value({'favorite_number': 10}, {}), 10)
        with self.assertRaises(FieldNotImplemented):
            f.value({'favorite_number': 9}, {})

    def test_type_error(self):
        colors = enum.make('Colors of the rainbow', {
            'Red': 'the color you bleed',
            'Orange': 'tasty fruit',
            'Yellow': 'like the sun',
            'Green': 'leaves in the summer',
            'Blue': 'cool ocean breeze',
            'Indigo': 'purple',
            'Violet': 'also purple',
        })
        fields = [
            StringField('one', lambda s, i, v: False),
            BooleanField('two', lambda s, i, v: "hello"),
            IntegerField('three', lambda s, i, v: 400.14),
            FloatField('four', lambda s, i, v: 900),
            EnumField('five', colors, lambda s, i, v: "again?"),
        ]
        for f in fields:
            f.__form_init__(self.form)

            with self.assertRaises(TypeError):
                f.value({}, {})

    def test_string_field(self):
        f = StringField('name', lambda s, i, v: "your name is: " + i['name'])
        f.__form_init__(self.form)

        self.assertEqual(f.value({'name': 'Bob'}, {}), "your name is: Bob")

        for value in ['John Smith ', '', '839']:
            string = f.to_string(value)
            value2 = f.from_string(string)
            self.assertEqual(value, string)
            self.assertEqual(value, value2)

    def test_boolean_field(self):
        f = BooleanField('has_name', lambda s, i, v: len(i['name']) > 0)
        f.__form_init__(self.form)

        self.assertTrue(f.value({'name': 'Bob'}, {}))
        self.assertFalse(f.value({'name': ''}, {}))

        for value in [True, False]:
            string = f.to_string(value)
            value2 = f.from_string(string)
            self.assertEqual(value, value2)

    def test_integer_field(self):
        f = IntegerField('name_length', lambda s, i, v: len(i['name']))
        f.__form_init__(self.form)

        self.assertEqual(f.value({'name': 'Bob'}, {}), 3)
        self.assertEqual(f.value({'name': ''}, {}), 0)

        for value in [98317, 0, -8, 10000]:
            string = f.to_string(value)
            value2 = f.from_string(string)
            self.assertEqual(string, str(value))
            self.assertEqual(value, value2)

    def test_float_field(self):
        f = FloatField('bytwo', lambda s, i, v: i['number'] / 2.0, places=5)
        f.__form_init__(self.form)

        self.assertAlmostEqual(f.value({'number': 2}, {}), 1, places=10)
        self.assertAlmostEqual(f.value({'number': -515}, {}), -257.5, places=10)
        self.assertAlmostEqual(f.value({'number': 2.6666666666}, {}), 1.33333, places=10)

        for v, s, v2 in [(float(98317), "98317.00000", 98317),
                         (-0.0091, "-0.00910", -0.0091),
                         (-8.888899, "-8.88890", -8.8889),
                         (0.0001, "0.00010", 0.0001),
                         (10000, "10000.00000", 10000)]:
            string = f.to_string(v)
            value2 = f.from_string(string)
            self.assertEqual(string, s)
            self.assertAlmostEqual(v2, value2, places=10)

    def test_float_field_rounding(self):

        f = FloatField('bytwo', lambda s, i, v: i['number'] / 2.0, places=2)
        f.__form_init__(self.form)

        self.assertAlmostEqual(f.value({'number': 2}, {}), 1, places=5)
        self.assertAlmostEqual(f.value({'number': -515}, {}), -257.50, places=5)
        self.assertAlmostEqual(f.value({'number': 2.6666666666}, {}), 1.33, places=5)
        self.assertAlmostEqual(f.value({'number': 98.999999}, {}), 49.5, places=5)

        for v, s, v2 in [(float(98317), "98317.00", 98317),
                         (.238, "0.24", 0.24),
                         (-8.8888, "-8.89", -8.89),
                         (0.0001, "0.00", 0),
                         (10000, "10000.00", 10000),
                         (98.2, "98.20", 98.2),
                         (-0.0091, "-0.01", -0.01)]:
            string = f.to_string(v)
            value2 = f.from_string(string)
            self.assertEqual(string, s)
            self.assertAlmostEqual(v2, value2, places=10)



from habutax.enum import filing_status_2021 as filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *

class Form1040RecoveryRebateCreditWkst(Form):
    form_name = "1040_recovery_rebate_credit_wkst"
    tax_year = 2021
    description = "Recovery Rebate Credit Worksheet (Form 1040)"
    long_description = "Line 30"
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('ssn_before_due_date', description="Does your 2021 return include a social security number that was issued on or before the due date of your 2021 return (including extensions) for you and, if filing a joint return, your spouse?"),
            BooleanInput('armed_forces', description="Was at least one of you or your spouse a member of the U.S. Armed Forces at any time during 2021, and does at least one of you have a social security number that was issued on or before the due date of your 2021 return (including extensions)?"),
            BooleanInput('either_ssn_before_due_date', description="Does one of you or your spouse have a social security number that was issued on or before the due date of your 2021 return (including extensions)?"),
            IntegerInput('dependents_ssn_before_due_date', description="How many dependents listed in the Dependents section on page 1 of Form 1040 or 1040-SR have a social security number that was issued on or before the due date of your 2021 return (including extensions) or an adoption taxpayer identification number?"),
            FloatInput('eip_3_amount', description="Enter the amount, if any, of EIP 3 that was issued to you. If filing a joint return, include the amount, if any, of your spouse’s EIP 3. You may refer to Notice 1444-C or your tax account information at IRS.gov/Account for the amount to enter here. Don’t include on line 13 any amount you received but later returned to the IRS."),
        ]

        def can_take_credit(self, i, v):
            if v['1']:
                return False
            if not v['2']:
                if i['1040.filing_status'] is filing_status.MarriedFilingJointly:
                    if not v['3'] and not v['4'] and not v['5']:
                        return False
                elif not v['5']:
                    return False
            if v['10_checkbox']:
                return False
            return True

        def line_6(self, i, v):
            amount = 1400.0
            if i['1040.filing_status'] is filing_status.MarriedFilingJointly and (v['2'] or v['3']):
                amount = 2800.0
            if not v['2']:
                if i['1040.filing_status'] is filing_status.MarriedFilingJointly:
                    if not v['3'] and not v['4'] and v['5']:
                        amount = 0.0
                elif v['5']:
                    amount = 0.0
            return amount

        def line_9_checkbox(self, i, v):
            amount = 0
            if i['1040.filing_status'] in [filing_status.Single, filing_status.MarriedFilingSeparately]:
                amount = 75000.0
            elif i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingWidowWidower]:
                amount = 150000.0
            elif i['1040.filing_status'] == filing_status.HeadOfHousehold:
                amount = 112.500
            else:
                self.not_implemented()
            return v['1040.11'] > amount

        def line_10_amount(self, i, v):
            if i['1040.filing_status'] in [filing_status.Single, filing_status.MarriedFilingSeparately]:
                return 80000.0
            elif i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingWidowWidower]:
                return 160000.0
            elif i['1040.filing_status'] == filing_status.HeadOfHousehold:
                return 120000.0
            else:
                self.not_implemented()

        def line_10_checkbox(self, i, v):
            return v['9'] > line_10_amount(self, i, v)

        def line_11(self, i, v):
            if i['1040.filing_status'] in [filing_status.Single, filing_status.MarriedFilingSeparately]:
                amount = 5000.0
            elif i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingWidowWidower]:
                amount = 10000.0
            elif i['1040.filing_status'] == filing_status.HeadOfHousehold:
                amount = 75000.0
            else:
                self.not_implemented()
            return v['10'] / amount

        optional_fields = [
            BooleanField('1', lambda s, i, v: False if i['1040.filing_status'] is filing_status.MarriedFilingJointly else i['1040.claimed_as_dependent']),
            BooleanField('2', lambda s, i, v: i['ssn_before_due_date']),
            BooleanField('3', lambda s, i, v: i['armed_forces']),
            BooleanField('4', lambda s, i, v: i['either_ssn_before_due_date']),
            BooleanField('5', lambda s, i, v: i['dependents_ssn_before_due_date'] > 0),
            FloatField('6', line_6),
            FloatField('7', lambda s, i, v: 1400.0 * i['dependents_ssn_before_due_date']),
            FloatField('8', lambda s, i, v: v['6'] + v['7']),
            BooleanField('9_checkbox', line_9_checkbox),
            FloatField('9', lambda s, i, v: v['1040.11']),
            BooleanField('10_checkbox', line_10_checkbox),
            FloatField('10', lambda s, i, v: s.not_implemented() if v['9'] > line_10_amount(s, i, v) else line_10_amount(s, i, v) - v['9']),
            FloatField('11', line_11),
            FloatField('12', lambda s, i, v: (v['8'] * v['11']) if v['9_checkbox'] else v['8']),
            FloatField('13', lambda s, i, v: i['eip_3_amount']),
            FloatField('14', lambda s, i, v: max(0.0, v['12'] - v['13'])),
            FloatField('credit', lambda s, i, v: v['14'] if can_take_credit(s, i, v) else None),
        ]

        super().__init__(__class__, inputs, [], optional_fields, **kwargs)

    def needs_filing(self, values):
        return False



import habutax.enum as enum
from habutax.form import InputForm, Jurisdiction
from habutax.inputs import *

class Form1099DIV(InputForm):
    form_name = "1099-div"
    tax_year = 2023
    description = "Form 1099-DIV"
    long_description = "Dividends and Distributions"
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        inputs = [
            EnumInput('belongs_to', enum.taxpayer_spouse_or_both, description="To whom does this 1099-DIV belong?"),
            StringInput('payer', description="Who paid you these dividends?"),
            StringInput('recipient', description="To whom was this dividend paid? (you or your spouse's name)"),
            FloatInput('box_1a', description="Total ordinary dividends"),
            FloatInput('box_1b', description="Qualified dividends"),
            FloatInput('box_2a', description="Total capital gain distr."),
            FloatInput('box_2b', description="Unrecap. Sec. 1250 gain"),
            FloatInput('box_2c', description="Section 1202 gain"),
            FloatInput('box_2d', description="Collectibles (28%) gain"),
            FloatInput('box_2e', description="Section 897 ordinary dividends"),
            FloatInput('box_2f', description="Section 897 capital gain"),
            FloatInput('box_3', description="Nondividend distributions"),
            FloatInput('box_4', description="Federal income tax withheld"),
            FloatInput('box_5', description="Section 199A dividends"),
            FloatInput('box_6', description="Investment expenses"),
            FloatInput('box_7', description="Foreign tax paid"),
            StringInput('box_8', description="Foreign country or US possesion"),
            FloatInput('box_9', description="Cash liquidation distributions"),
            FloatInput('box_10', description="Noncash liquidation distributions"),
            BooleanInput('box_11', description="FATCA filing requirement box checked"),
            FloatInput('box_12', description="Exempt-interest dividend"),
            FloatInput('box_13', description="Specified private activity bond interest dividends"),
            EnumInput('box_14_1', enum.us_states, allow_empty=True, description="State #1"),
            StringInput('box_15_1', description="State #1 identification no."),
            FloatInput('box_16_1', description="State #1 tax withheld"),
            EnumInput('box_14_2', enum.us_states, allow_empty=True, description="State #2"),
            StringInput('box_15_2', description="State #2 identification no."),
            FloatInput('box_16_2', description="State #2 tax withheld"),
        ]

        super().__init__(__class__, inputs, **kwargs)



import configparser
from collections.abc import MutableMapping

class UnmetDependency(Exception):
    def __init__(self, dependency_name, message_fmt="Unmet dependency: {dependency}"):
        self.dependency = dependency_name
        self.message = message_fmt.format(dependency=dependency_name)
        super().__init__(self.message)

class ValueStore(MutableMapping):
    def __init__(self, *args, **kwargs):
        self.values = dict()
        self.update(dict(*args, **kwargs))

    def __getitem__(self, key):
        try:
            return self.values[key]
        except KeyError as ke:
            raise UnmetDependency(key) from ke

    def __setitem__(self, key, value):
        self.values[key] = value

    def __delitem__(self, key):
        del self.values[key]

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def to_config(self, field_map):
        config = configparser.ConfigParser()
        for key, value in self.values.items():
            field = field_map[key]
            form_name, field_name = key.split('.')
            if form_name not in config:
                config[form_name] = {}
            config[form_name][field_name] = field.to_string(value)
        return config



from habutax.enum import filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *

from habutax.forms.ty2022.f1040_figure_tax import figure_tax

class Form1040QualDivCapGainTaxWkst(Form):
    form_name = "1040_qualdiv_capgain_tax_wkst"
    tax_year = 2022
    description = "Qualified Dividends and Capital Gain Tax Worksheet (Form 1040)"
    long_description = "Line 16"
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('form_2555', description="Do you need to fill out Form 2555 (relating to foreign earned income)?"),
        ]

        def line_6(self, i, v):
            if i['1040.filing_status'] in [filing_status.Single, filing_status.MarriedFilingSeparately]:
                return 41675.0
            elif i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingSurvivingSpouse]:
                return 83350.0
            elif i['1040.filing_status'] == filing_status.HeadOfHousehold:
                return 55800.0
            else:
                self.not_implemented()

        def line_13(self, i, v):
            if i['1040.filing_status'] is filing_status.Single:
                return 459750.00
            if i['1040.filing_status'] is filing_status.MarriedFilingSeparately:
                return 258600.00
            elif i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingSurvivingSpouse]:
                return 517200.00
            elif i['1040.filing_status'] == filing_status.HeadOfHousehold:
                return 488500.00
            else:
                self.not_implemented()

        optional_fields = [
            FloatField('1', lambda s, i, v: s.not_implemented() if i['form_2555'] else v['1040.15']),
            FloatField('2', lambda s, i, v: s.not_implemented() if i['form_2555'] else v['1040.3a']),
            FloatField('3', lambda s, i, v: s.not_implemented() if i['1040.schedule_d_required'] else v['1040.7']),
            FloatField('4', lambda s, i, v: v['2'] + v['3']),
            FloatField('5', lambda s, i, v: max(0.0, v['1'] - v['4'])),
            FloatField('6', line_6),
            FloatField('7', lambda s, i, v: min(v['1'], v['6'])),
            FloatField('8', lambda s, i, v: min(v['5'], v['7'])),
            FloatField('9', lambda s, i, v: v['7'] - v['8']),
            FloatField('10', lambda s, i, v: min(v['1'], v['4'])),
            FloatField('11', lambda s, i, v: v['9']),
            FloatField('12', lambda s, i, v: v['10'] - v['11']),
            FloatField('13', line_13),
            FloatField('14', lambda s, i, v: min(v['1'], v['13'])),
            FloatField('15', lambda s, i, v: v['5'] + v['9']),
            FloatField('16', lambda s, i, v: max(0.0, v['14'] - v['15'])),
            FloatField('17', lambda s, i, v: min(v['12'], v['16'])),
            FloatField('18', lambda s, i, v: v['17'] * 0.15),
            FloatField('19', lambda s, i, v: v['9'] + v['17']),
            FloatField('20', lambda s, i, v: v['10'] - v['19']),
            FloatField('21', lambda s, i, v: v['20'] * 0.20),
            FloatField('22', lambda s, i, v: figure_tax(v['5'], i['1040.filing_status'])),
            FloatField('23', lambda s, i, v: v['18'] + v['21'] + v['22']),
            FloatField('24', lambda s, i, v: figure_tax(v['1'], i['1040.filing_status'])),
            FloatField('25', lambda s, i, v: min(v['23'], v['24'])),
        ]

        super().__init__(__class__, inputs, [], optional_fields, **kwargs)

    def needs_filing(self, values):
        return False



from enum import Enum, auto
import copy
import os
import re
import unittest

from . import context


class DocumentationTestCase(unittest.TestCase):
    def find_doc_code(self, documentation_filename, code_filename):

        escaped_code_filename = re.escape(code_filename)
        code_listing = re.compile(f'^example code: {escaped_code_filename}$')
        start = re.compile('^```python$')
        stop = re.compile('^```$')

        class State(Enum):
            SEARCHING = auto()
            CONFIRMING = auto()
            READING = auto()

        code = ""
        full_doc_filename = os.path.join(context.habutax_basedir, 'doc', documentation_filename)
        with open(full_doc_filename) as doc:
            state = State.SEARCHING
            for line in doc:
                if state == State.READING:
                    if stop.match(line):
                        self.assertTrue(len(code) > 0)
                        return code
                    else:
                        code += line
                elif state == State.CONFIRMING:
                    if start.match(line):
                        state = State.READING
                    else:
                        state = State.SEARCHING
                elif code_listing.match(line):
                    state = State.CONFIRMING

        self.assertTrue(False)

    def find_compile_doc_code(self, documentation_filename, code_filename):
        code = self.find_doc_code(documentation_filename, code_filename)
        return compile(code, f'doc/{documentation_filename} example code: {code_filename}', 'exec')

    def setUp(self):
        """Because the tests in this class exec() code examples, which causes
        pollution of the global space (via imports or definitions of classes,
        etc.), save as much of the current state of the globals as we can to
        restore it later"""
        self._saved_globals = {}
        self._keep_globals = set()

        for k, v in globals().items():
            try:
                self._saved_globals[k] = copy.deepcopy(globals()[k])
            except:
                self._keep_globals.add(k)

    def tearDown(self):
        """Restore global state as it was saved by setUp() as much as we can"""
        for k in list(globals().keys()):
            if k not in self._saved_globals and k not in self._keep_globals:
                del globals()[k]
            elif k in self._saved_globals:
                globals()[k] = self._saved_globals[k]

        del self._saved_globals
        del self._keep_globals

    def test_form0000_example(self):
        f0000 = self.find_compile_doc_code('adding_modifying_forms.md', 'f0000.py')

        # Defines class Form0000
        exec(f0000, globals())

        f0 = Form0000()
        self.assertEqual(f0.name(), "0000")

        self.assertEqual(len(f0.inputs()), 3)
        self.assertEqual(len(f0.required_fields()), 0)
        self.assertEqual(len(f0.fields()), 2)

        self.assertEqual(f0.inputs()[0].name(), "0000.base_tax")
        self.assertEqual(f0.fields()[1].name(), "0000.2")

        # Test with 'naughty' input
        naughty_inputs = {
            'base_tax': 8429.3,
            'nice': False,
            'naughty_tax': 83.50,
        }
        naughty_fields = {
            '1': f0.fields()[0].value(naughty_inputs, {}),
        }
        naughty_fields['2'] = f0.fields()[1].value(naughty_inputs, naughty_fields)

        self.assertAlmostEqual(naughty_fields['2'], 8512.80, places=2)

        # Test with 'nice' input
        nice_inputs = {
            'base_tax': 8429.3,
            'nice': True,
        }
        nice_fields = {
            '1': f0.fields()[0].value(nice_inputs, {}),
        }
        nice_fields['2'] = f0.fields()[1].value(nice_inputs, nice_fields)

        self.assertAlmostEqual(nice_fields['2'], 8429.3, places=2)

    def test_formsomething_example(self):
        fsomething = self.find_compile_doc_code('adding_modifying_forms.md', 'fsomething.py')

        # Defines class FormSomething
        exec(fsomething, globals())

        f0 = FormSomething()
        self.assertEqual(len(f0.inputs()), 3)
        self.assertEqual(len(f0.required_fields()), 3)
        self.assertEqual(len(f0.fields()), 3)

        inputs = {
            'name': "Monty",
            'ssn': "000-00-0000",
            'eye_color': f0.EYES.brown,
        }

        values = {}
        for field in f0.required_fields():
            values[field.base_name()] = field.value(inputs, values)

        self.assertEqual(values['name'], "Monty")
        self.assertEqual(values['ssn'], "000-00-0000")
        self.assertEqual(values['eye_color'], f0.EYES.brown)

    def test_fthresholds_example(self):
        fthresholds = self.find_compile_doc_code('adding_modifying_forms.md', 'fthresholds.py')

        # Defines class FormThresholds
        exec(fthresholds, globals())

        ft = FormThresholds()
        self.assertEqual(len(ft.inputs()), 2)
        self.assertEqual(len(ft.required_fields()), 2)
        self.assertEqual(len(ft.fields()), 2)

        inputs = {
            'base_tax': 121.3,
            'eye_color': ft.EYES.brown,
        }

        values = {}
        for field in ft.required_fields():
            values[field.base_name()] = field.value(inputs, values)

        self.assertAlmostEqual(values['1'], 500.0, places=2)
        self.assertAlmostEqual(values['2'], 520.0, places=2)



from .f1040 import Form1040
from .f1040_qualdiv_capgain_tax_wkst import Form1040QualDivCapGainTaxWkst
from .f1040_s1 import Form1040S1
from .f1040_s2_need6251 import Form1040S2Need6251
from .f1040_s3 import Form1040S3
from .f1040_sa import Form1040SA
from .f1040_sb import Form1040SB
from .f1040_s8812 import Form1040S8812
from .f1098 import Form1098
from .f1099_div import Form1099DIV
from .f1099_g import Form1099G
from .f1099_int import Form1099INT
from .f1099_r import Form1099R
from .f8606 import Form8606
from .f8959 import Form8959
from .f8995 import Form8995
from .fw_2 import FormW2
from .f8889 import Form8889
from .fnc_d_400 import FormNCD400
from .fnc_d_400_child_deduction_wkst import FormNCD400ChildDeductionWkst
from .fnc_d_400_consumer_use_tax_wkst import FormNCD400ConsumerUseTaxWkst
from .fnc_d_400_ss import FormNCD400SS
from .fnc_d_400_sa import FormNCD400SA

available_forms = [
    Form1040,
    Form1040QualDivCapGainTaxWkst,
    Form1040S1,
    Form1040S2Need6251,
    Form1040S3,
    Form1040SA,
    Form1040SB,
    Form1040S8812,
    Form1098,
    Form1099DIV,
    Form1099G,
    Form1099INT,
    Form1099R,
    Form8606,
    Form8959,
    Form8995,
    FormW2,
    Form8889,
    FormNCD400,
    FormNCD400ChildDeductionWkst,
    FormNCD400ConsumerUseTaxWkst,
    FormNCD400SS,
    FormNCD400SA,
]



from ..test_form import FormTestCase

import habutax.forms.ty2022 as ty2022


class Form1040SimpleTestCase(FormTestCase):
    def setUp(self):
        self.fixture_setup(ty2022.available_forms, input_fixture='tests/ty2022/fixtures/f1040_one_w2.habutax')

    def test_simple_1040(self):
        solution = self.assertSolve(['1040'])

        self.assertIn('1040', solution)
        self.assertIn('w-2:0', solution)

        self.assertDollarsEqual(solution, '1040.1a', 100000) # W-2 income
        self.assertDollarsEqual(solution, '1040.11', 100000) # AGI
        self.assertDollarsEqual(solution, '1040.12', 12950)  # standard deduction
        self.assertDollarsEqual(solution, '1040.15', 100000 - 12950) # taxable income
        self.assertDollarsEqual(solution, '1040.16', 14774) # tax
        self.assertDollarsEqual(solution, '1040.24', 14774) # total tax
        self.assertDollarsEqual(solution, '1040.25a', 16551.90) # W-2 withholding
        self.assertDollarsEqual(solution, '1040.25d', 16551.90) # total withholding
        self.assertDollarsEqual(solution, '1040.33', 16551.90) # total payments
        self.assertDollarsEqual(solution, '1040.34', 1777.90) # overpayment
        self.assertDollarsEqual(solution, '1040.35a', 1777.90) # refund
        self.assertDollarsEqual(solution, '1040.37', 0) # amount you owe
        self.assertDollarsEqual(solution, '1040.38', 0) # tax penalty



import unittest
from configparser import ConfigParser

from habutax.inputs import *
from habutax.fields import *
from habutax.form import *
from habutax.solver import Solver, DependencyTracker, sort_keys


class TestForm(Form):
    form_name = "test"
    tax_year = 1970

    def __init__(self, **kwargs):
        test_inputs = [
            IntegerInput('bar', description="I don't know"),
        ]
        test_fields = [
            IntegerField('foo', lambda s, i, v: i['bar']),
            IntegerField('something', lambda s, i, v: i['bar']),
            IntegerField('else', lambda s, i, v: v['something'] - i['bar']),
        ]
        super().__init__(__class__, test_inputs, test_fields, [], **kwargs)


class DependencyTrackerTestCase(unittest.TestCase):
    def setUp(self):
        self.form = TestForm()
        self.dep = DependencyTracker()

    def test_add_unmet(self):
        self.dep.add_unmet('test.bar', self.form.required_fields()[0])

        self.assertTrue(self.dep.has_unmet())
        self.assertFalse(self.dep.has_met())

        unmet_dependencies = self.dep.unmet_dependencies()
        self.assertEqual(len(unmet_dependencies), 1)
        self.assertEqual(unmet_dependencies[0], 'test.bar')

        unmet_dependents = self.dep.unmet_dependents('test.bar')
        self.assertEqual(len(unmet_dependents), 1)
        self.assertEqual(unmet_dependents[0].name(), 'test.foo')

        self.dep.add_unmet('test.bar', self.form.required_fields()[1])
        self.dep.add_unmet('test.bar', self.form.required_fields()[2])

        self.assertTrue(self.dep.has_unmet())
        self.assertFalse(self.dep.has_met())

        unmet_dependencies = self.dep.unmet_dependencies()
        self.assertEqual(len(unmet_dependencies), 1)
        self.assertEqual(unmet_dependencies[0], 'test.bar')

        unmet_dependents = self.dep.unmet_dependents('test.bar')
        self.assertEqual(len(unmet_dependents), 3)
        unmet_dependent_names = [ud.name() for ud in unmet_dependents]
        self.assertIn('test.foo', unmet_dependent_names)
        self.assertIn('test.something', unmet_dependent_names)
        self.assertIn('test.else', unmet_dependent_names)

    def test_meet(self):
        self.dep.add_unmet('test.bar', self.form.required_fields()[0])
        self.dep.add_unmet('test.bar', self.form.required_fields()[1])
        self.dep.add_unmet('test.bar', self.form.required_fields()[2])

        self.assertTrue(self.dep.has_unmet())
        self.assertFalse(self.dep.has_met())

        self.dep.meet('test.bar')

        self.assertFalse(self.dep.has_unmet())
        self.assertTrue(self.dep.has_met())

    def test_met_dependents(self):
        self.dep.add_unmet('test.bar', self.form.required_fields()[0])
        self.dep.add_unmet('test.bar', self.form.required_fields()[1])
        self.dep.add_unmet('test.bar', self.form.required_fields()[2])

        self.assertEqual(len(list(self.dep.met_dependents())), 0)

        self.dep.meet('test.bar')

        self.assertFalse(self.dep.has_unmet())
        self.assertTrue(self.dep.has_met())

        met_dependent_names = [f.name() for f in self.dep.met_dependents()]
        self.assertEqual(len(met_dependent_names), 3)

        self.assertIn('test.foo', met_dependent_names)
        self.assertIn('test.something', met_dependent_names)
        self.assertIn('test.else', met_dependent_names)

        self.assertFalse(self.dep.has_unmet())
        self.assertFalse(self.dep.has_met())


class SortOrderTestCase(unittest.TestCase):
    def test_sort_keys_same_form(self):
        original = [
            '1040.10',
            '1040.829',
            '1040.1a',
            '1040.8a8',
            '1040.1',
        ]
        expected = [
            '1040.1',
            '1040.1a',
            '1040.8a8',
            '1040.10',
            '1040.829',
        ]
        self.assertListEqual(sorted(original, key=sort_keys), expected)

    def test_sort_keys_no_form(self):
        original = [
            '2',
            '5_wkst_to_do_thing_10',
            '20',
            '5_wkst_to_do_thing_1',
            '1a',
            '5_wkst_to_do_thing_2',
            '1',
            '5_wkst_to_do_thing_9a',
            '1c',
            '5_wkst_to_do_thing_90',
        ]
        expected = [
            '1',
            '1a',
            '1c',
            '2',
            '5_wkst_to_do_thing_1',
            '5_wkst_to_do_thing_2',
            '5_wkst_to_do_thing_9a',
            '5_wkst_to_do_thing_10',
            '5_wkst_to_do_thing_90',
            '20',
        ]
        self.assertListEqual(expected, sorted(original, key=sort_keys))

    def test_sort_keys_different_forms(self):
        original = [
            '1040_sa.10',
            '8995.5b',
            '1040_sb.829',
            '1040_s8812.1a',
            'w-2.box_1',
            '1040.first_name',
            '8959.5b',
        ]
        expected = [
            '1040.first_name',
            '1040_s8812.1a',
            '1040_sa.10',
            '1040_sb.829',
            '8959.5b',
            '8995.5b',
            'w-2.box_1',
        ]
        self.assertListEqual(sorted(original, key=sort_keys), expected)

    def test_sort_keys_mixed_form_presence(self):
        original = [
            '1040_sb.829',
            '10',
            '5b',
            '1040.first_name',
            '51',
            '1040_s8812.1a',
            'box_1',
            '5a',
            '8995.5b',
        ]
        expected = [
            '5a',
            '5b',
            '10',
            '51',
            'box_1',
            '1040.first_name',
            '1040_s8812.1a',
            '1040_sb.829',
            '8995.5b',
        ]
        self.assertListEqual(sorted(original, key=sort_keys), expected)

    def test_sort_keys_mixed_types(self):
        test_form = TestForm()
        original = [
            "test.9a",
            test_form.inputs()[0],
            "z.foobar",
            test_form.fields()[0],
            "test.c",
            test_form.fields()[1],
            "test.z00",
            "000.1",
            test_form.fields()[2],
        ]
        expected = [
            "000.1",
            "test.9a",
            test_form.inputs()[0], # test.bar
            "test.c",
            test_form.fields()[2], # test.else
            test_form.fields()[0], # test.foo
            test_form.fields()[1], # test.something
            "test.z00",
            "z.foobar",
        ]
        self.assertListEqual(sorted(original, key=sort_keys), expected)


class SolverTestCase(unittest.TestCase):
    def setUp(self):
        self.config = ConfigParser()
        self.inputs = InputStore(self.config)
        self.solver = Solver(self.inputs, [TestForm])

    def test_basic_form(self):
        self.config['test'] = {'bar': '5'}
        solved = self.solver.solve(['test'])

        self.assertTrue(solved)
        solution = self.solver.solution()
        self.assertIn('test', solution)

        for field, value in (('foo', '5'), ('something', '5'), ('else', '0')):
            self.assertIn(field, solution['test'])
            self.assertEqual(solution['test'][field], value)

    def test_basic_form_missing_input(self):
        solved = self.solver.solve(['test'])

        self.assertFalse(solved)
        solution = self.solver.solution()
        self.assertNotIn('test', solution)

        self.assertEqual(len(self.solver.unimplemented_fields()), 0)

        dep = self.solver.unmet_input_dependencies()
        self.assertEqual(len(dep), 1)
        self.assertIn('test.bar', dep)
        self.assertEqual(len(dep['test.bar']), 2)
        self.assertIn('test.foo', dep['test.bar'])
        self.assertIn('test.something', dep['test.bar'])

        self.assertEqual(len(self.solver.unmet_field_dependencies()), 1)



from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *

class FormNCD400ConsumerUseTaxWkst(Form):
    form_name = "nc_d-400_consumer_use_tax_wkst"
    tax_year = 2023
    description = "Consumer Use Tax Worksheet (NC Form D-400)"
    long_description = "Line 18"
    jurisdiction = Jurisdiction.NC

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('full_records', description="Do you have records of out-of-state (NC) purchases your or your spouse made for tax year 2023?"),
            FloatInput('out_of_state_purchases', description="Enter the total amount of out-of-state purchases, including delivery charges, for 1/1/2023 through 12/31/2023"),
            FloatInput('county_tax_pct', description="Enter 7.5% (.075) in Durham and Orange Counties; 7.25% (.0725) in Mecklenburg and Wake Counties; 7%(.07) in Alexander, Alleghany, Anson, Ashe, Bertie, Buncombe, Cabarrus, Catawba, Chatham, Cherokee, Clay, Cumberland, Davidson, Duplin, Edgecombe, Forsyth, Gaston, Graham, Greene, Halifax, Harnett, Haywood, Hertford, Jackson, Jones, Lee, Lincoln, Madison, Martin, Montgomery, Moore, New Hanover, Onslow, Pasquotank, Pitt, Randolph, Robeson, Rockingham, Rowan, Rutherford, Sampson, Stanly, Surry, Swain, and Wilkes Counties; and 6.75% (.0675) in all other counties"),
            FloatInput('other_state_sales_tax', description="Enter the amount of sales tax legally and properly paid to another state or North Carolina on your out-of-state purchases."),
        ]

        def estimate(self, i, v):
            use_tax_table = [
                (2200, 1),
                (3700, 2),
                (5200, 3),
                (6700, 4),
                (8100, 5),
                (9600, 6),
                (11100, 7),
                (12600, 8),
                (14100, 9),
                (15600, 10),
                (17000, 11),
                (18500, 12),
                (20000, 13),
                (21500, 14),
                (23000, 15),
                (24400, 16),
                (25900, 17),
                (27400, 18),
                (28900, 19),
                (30400, 20),
                (31900, 21),
                (33300, 22),
                (34800, 23),
                (36300, 24),
                (37800, 25),
                (39300, 26),
                (40700, 27),
                (42200, 28),
                (43700, 29),
                (45200, 30),
            ]
            for max_taxable_income, use_tax in use_tax_table:
                if v['nc_d-400.14'] < max_taxable_income:
                    return float(use_tax)
            assert v['nc_d-400.14'] >= 45200
            return v['nc_d-400.14'] * 0.000675

        optional_fields = [
            FloatField('estimate', estimate, places=0),
            FloatField('1', lambda s, i, v: i['out_of_state_purchases'], places=0),
            FloatField('2', lambda s, i, v: i['county_tax_pct'] * v['1'], places=0),
            FloatField('3', lambda s, i, v: min(i['other_state_sales_tax'], v['2']), places=0),
            FloatField('4', lambda s, i, v: round(v['2'] - v['3'], 0), places=0),
            FloatField('consumer_use_tax', lambda s, i, v: v['4'] if i['full_records'] else v['estimate'], places=0),
        ]

        required_fields = [
        ]

        super().__init__(__class__, inputs, required_fields, optional_fields,
                         **kwargs)

    def needs_filing(self, values):
        return False



from math import ceil
import os

from habutax.enum import filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class Form1040S8812(Form):
    form_name = "1040_s8812"
    tax_year = 2022
    description = "Schedule 8812 (Form 1040)"
    long_description = "Credits for Qualifying Children and Other Dependents"
    jurisdiction = Jurisdiction.US
    sequence_no = 47

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('pr_income_forms_2555_4563', description="Did you exclude any income from Puerto Rico, or require Forms 2555 or 2563?"),
            IntegerInput('number_under_17', description="How many of your dependents eligible for the child tax credit with the required social security number were under age 17 with the required social security number at the end of 2022?"),
            BooleanInput('credit_limit_ws_b_maybe_needed', description='Are you using forms 5695, 8396, 8839, 8859 to report credits?'),
        ]

        def line_10(self, i, v):
            res = v['3'] - v['9']
            if res <= 0.001:
                return 0.0
            return ceil(res / 1000.0) * 1000.0

        def line_27(self, i, v):
            if not v['8_gt_11']:
                return 0.0
            if v['16a'] < 0.001:
                return 0.0
            if v['16b'] < 0.01:
                return 0.0
            return self.not_implemented("Schedule 8812 has not implemented the 'Additional Child Tax Credit for All Filers'")

        optional_fields = [
            # Part I-A: Child Tax Credit and Credit for Other Dependents
            FloatField('1', lambda s, i, v: v['1040.11']),
            FloatField('2a', lambda s, i, v: s.not_implemented() if i['pr_income_forms_2555_4563'] else None),
            FloatField('2b', lambda s, i, v: s.not_implemented() if i['pr_income_forms_2555_4563'] else None),
            FloatField('2c', lambda s, i, v: s.not_implemented() if i['pr_income_forms_2555_4563'] else None),
            FloatField('2d', lambda s, i, v: v['2a'] + v['2b'] + v['2c']),
            FloatField('3', lambda s, i, v: v['1'] + v['2d']),
            IntegerField('4', lambda s, i, v: i['number_under_17']),
            FloatField('5', lambda s, i, v: v['4'] * 2000.0),
            IntegerField('6', lambda s, i, v: sum([v[f'1040.dependent_{n}_odc'] for n in range(i['1040.number_dependents'])])),
            FloatField('7', lambda s, i, v: v['6'] * 500.0),
            FloatField('8', lambda s, i, v: v['5'] + v['7']),
            FloatField('9', lambda s, i, v: 400000.0 if i['1040.filing_status'] is filing_status.MarriedFilingJointly else 200000.0),
            FloatField('10', line_10),
            FloatField('11', lambda s, i, v: v['10'] * 0.05),
            BooleanField('8_gt_11', lambda s, i, v: v['8'] > v['11']),
            FloatField('12', lambda s, i, v: v['8'] - v['11'] if v['8_gt_11'] else None),
            FloatField('13', lambda s, i, v: v['clwkst_a_5']),
            FloatField('14', lambda s, i, v: min(v['12'], v['13']) if v['8_gt_11'] else 0.0),

            # Note: we are omitting most of II-A and all of II-B. The
            # implementation of line_27 checks to make sure we're not hitting
            # these cases

            # Part II-A: Additional Child Tax Credit for All Filers
            FloatField('16a', lambda s, i, v: v['12'] - v['14']),
            FloatField('16b', lambda s, i, v: v['4'] * 1500.0),
            FloatField('17', lambda s, i, v: min(v['16a'], v['16b'])),

            # Part II-C: Additional Child Tax Credit
            FloatField('27', line_27),
        ]

        def line_clwkst_a_2(self, i, v):
            if not v['1040.need_schedule_3_part_i']:
                return None

            if i['1040_s3.residential_energy_credit']:
                self.not_implemented()

            amounts = sum([v[f'1040_s3.{l}'] for l in ['1', '2', '3', '4', '6d', '6e', '6f', '6l']])
            return None if amounts < 0.001 else amounts

        optional_fields += [
            # Credit Limit Worksheet A
            FloatField('clwkst_a_1', lambda s, i, v: v['1040.18']),
            FloatField('clwkst_a_2', line_clwkst_a_2),
            FloatField('clwkst_a_3', lambda s, i, v: v['clwkst_a_1'] - v['clwkst_a_2']),
            FloatField('clwkst_a_4', lambda s, i, v: s.not_implemented() if i['credit_limit_ws_b_maybe_needed'] else 0.0),
            FloatField('clwkst_a_5', lambda s, i, v: v['clwkst_a_3'] - v['clwkst_a_4']),
        ]

        required_fields = [
            BooleanField('dependents_match', lambda s, i, v: True if sum([v[f'1040.dependent_{n}_ctc'] for n in range(i['1040.number_dependents'])]) == v['4'] else s.not_implemented()), # TODO would be more user-friendly to be reported as some type of error rather than "not implemented"
        ]

        pdf_fields = [
            TextPDFField('topmostSubform[0].Page1[0].f1_1[0]', '1040.full_names'),
            TextPDFField('topmostSubform[0].Page1[0].f1_2[0]', '1040.you_ssn', max_length=11),
            TextPDFField('topmostSubform[0].Page1[0].f1_3[0]', '1'),
            TextPDFField('topmostSubform[0].Page1[0].f1_4[0]', '2a'),
            TextPDFField('topmostSubform[0].Page1[0].f1_5[0]', '2b'),
            TextPDFField('topmostSubform[0].Page1[0].f1_6[0]', '2c'),
            TextPDFField('topmostSubform[0].Page1[0].f1_7[0]', '2d'),
            TextPDFField('topmostSubform[0].Page1[0].f1_8[0]', '3'),
            TextPDFField('topmostSubform[0].Page1[0].f1_9[0]', '4'),
            TextPDFField('topmostSubform[0].Page1[0].f1_10[0]', '5'),
            TextPDFField('topmostSubform[0].Page1[0].Line6ReadOrder[0].f1_11[0]', '6'),
            TextPDFField('topmostSubform[0].Page1[0].f1_12[0]', '7'),
            TextPDFField('topmostSubform[0].Page1[0].f1_13[0]', '8'),
            TextPDFField('topmostSubform[0].Page1[0].f1_14[0]', '9'),
            TextPDFField('topmostSubform[0].Page1[0].f1_15[0]', '10'),
            TextPDFField('topmostSubform[0].Page1[0].f1_16[0]', '11'),
            TextPDFField('topmostSubform[0].Page1[0].f1_17[0]', '12'),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_1[0]', '8_gt_11', '1', lambda s, v, f: not v),
            ButtonPDFField('topmostSubform[0].Page1[0].c1_1[1]', '8_gt_11', '2', lambda s, v, f: v),
            TextPDFField('topmostSubform[0].Page1[0].f1_18[0]', '13'),
            TextPDFField('topmostSubform[0].Page1[0].f1_19[0]', '14'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_1[0]', 'do_not_want_to_claim_additional_ctc', '1'),
            TextPDFField('topmostSubform[0].Page2[0].f2_1[0]', '16a'),
            TextPDFField('topmostSubform[0].Page2[0].f2_2[0]', '4'), # number of qualifying children for 16b
            TextPDFField('topmostSubform[0].Page2[0].f2_3[0]', '16b'),
            TextPDFField('topmostSubform[0].Page2[0].f2_4[0]', '17'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_5[0]', '18a'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_6[0]', '18b'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_2[0]', 'unknown', '1'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_2[1]', 'unknown', '2'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_7[0]', '19'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_8[0]', '20'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_3[0]', 'unknown', '1'),
#            ButtonPDFField('topmostSubform[0].Page2[0].c2_3[1]', 'unknown', '2'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_9[0]', '21'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_10[0]', '22'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_11[0]', '23'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_12[0]', '24'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_13[0]', '25'),
#            TextPDFField('topmostSubform[0].Page2[0].f2_14[0]', '26'),
            TextPDFField('topmostSubform[0].Page2[0].f2_15[0]', '27'),
        ]

        pdf_file = os.path.join(os.path.dirname(__file__), 'f1040s8.pdf')
        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        return True



from ..test_form import FormTestCase

import habutax.forms.ty2021 as ty2021


class Form1040SimpleTestCase(FormTestCase):
    def setUp(self):
        self.fixture_setup(ty2021.available_forms, input_fixture='tests/ty2021/fixtures/f1040_one_w2.habutax')

    def test_simple_1040(self):
        solution = self.assertSolve(['1040'])

        self.assertIn('1040', solution)
        self.assertIn('w-2:0', solution)

        self.assertDollarsEqual(solution, '1040.1', 100000)
        self.assertDollarsEqual(solution, '1040.12c', 12550 + 300)
        self.assertDollarsEqual(solution, '1040.15', 100000 - (12550 + 300))
        self.assertDollarsEqual(solution, '1040.16', 14943)
        self.assertDollarsEqual(solution, '1040.25a', 16551.9)
        self.assertDollarsEqual(solution, '1040.25d', 16551.9)
        self.assertDollarsEqual(solution, '1040.33', 16551.9)
        self.assertDollarsEqual(solution, '1040.34', 1608.9)
        self.assertDollarsEqual(solution, '1040.35a', 1608.9)
        self.assertDollarsEqual(solution, '1040.37', 0)
        self.assertDollarsEqual(solution, '1040.38', 0)



import unittest

from habutax.forms.ty2021.f1040_figure_tax import figure_tax
from habutax import enum


class FigureTaxTestCase(unittest.TestCase):
    def test_figure_2021_1040_tax(self):
        scenarios = (
            (75047.08, enum.filing_status_2021.Single, 12254),
            (98325.84, enum.filing_status_2021.MarriedFilingJointly, 13129),
            (99999.99, enum.filing_status_2021.HeadOfHousehold, 16563.00),
            (100000.00, enum.filing_status_2021.HeadOfHousehold, 16569.00),
            (588939, enum.filing_status_2021.QualifyingWidowWidower, 155217.15),
            (588939, enum.filing_status_2021.MarriedFilingJointly, 155217.15),
            (1482.40, enum.filing_status_2021.MarriedFilingSeparately, 149),
            (8182983.94, enum.filing_status_2021.MarriedFilingSeparately, 2995965.31),
        )
        for taxable, status, expected_tax in scenarios:
            tax = figure_tax(taxable, status)
            self.assertAlmostEqual(expected_tax, tax, places=2)



from collections.abc import MutableMapping
import configparser
import re

class Input(object):
    def __init__(self, name, description=None):
        assert "." not in name
        self._name = name
        self._description = name if description is None else description

    def __form_init__(self, form):
        """Called when this Input is associated with a form instance"""
        self._form = form

    def section(self):
        return self._form.name()

    def base_name(self):
        """Return the name relative to the form it is in"""
        return self._name

    def name(self):
        return f'{self.section()}.{self.base_name()}'

    def help(self):
        return self._description

    def format_suggestion(self):
        raise NotImplementedError()

    def value(self, string):
        raise NotImplementedError()

    def valid(self, string):
        try:
            self.value(string)
        except ValueError:
            return False
        return True

class StringInput(Input):
    def format_suggestion(self):
        return ''

    def value(self, string):
        return string.strip()

class BooleanInput(Input):
    def format_suggestion(self):
        return 'Input one of y[es] or n[o]'

    def value(self, string):
        string = string.strip().lower()
        if string in ['true', 'yes', 'y', '1', 'on']:
            return True
        if string in ['false', 'no', 'n', '0', 'off']:
            return False
        raise ValueError(f'Invalid boolean value: {string}')

class IntegerInput(Input):
    def format_suggestion(self):
        return "Input must be an integer"

    def value(self, string):
        string = string.strip()
        if len(string) == 0:
            return 0
        return int(string)

class FloatInput(Input):
    def format_suggestion(self):
        return "Input must be a floating point number"

    def value(self, string):
        string = string.strip()
        if len(string) == 0:
            return 0.0
        return float(string)

class EnumInput(StringInput):
    def __init__(self, name, enum, allow_empty=False, description=None):
        """
        Create an instance to read/validate an input that can be one of a
        limited number of choices.
        """
        super().__init__(name, description=description)
        self.enum = enum
        self.allow_empty = allow_empty

    def format_suggestion(self):
        empty = "empty or " if self.allow_empty else ""
        suggestion = f'Input must be {empty}one of (in quotes):\n'
        for k, v in self.enum.__members__.items():
            suggestion += f' * "{k}": {v.value}\n'
        return suggestion.strip()

    def __form_init__(self, form):
        super().__form_init__(form)

    def __getattr__(self, attribute):
        return self.enum[attribute]

    def valid(self, string):
        try:
            string = super().value(string)
        except ValueError:
            return False

        if len(string.strip()) == 0 and self.allow_empty:
            return True

        try:
            self.enum[string]
        except KeyError as ke:
            return False
        return True

    def value(self, string):
        string = super().value(string)
        if len(string.strip()) == 0 and self.allow_empty:
            return None
        return self.enum[string]

class RegexInput(StringInput):
    def __init__(self, name, regex, description=""):
        self._regex_str = regex
        self._regex = re.compile(regex)
        super().__init__(name, description=description)

    def format_suggestion(self):
        return f'Input must match the regular expression {self._regex_str}'

    def valid(self, string):
        try:
            v = self.value(string)
        except ValueError:
            return False

        return bool(self._regex.match(v))

class SSNInput(StringInput):
    def value(self, string):
        v = super().value(string)
        return v.replace("-", "")

    def format_suggestion(self):
        return "Input should be in the form 123-45-6789"

    def valid(self, string):
        try:
            ssn = self.value(string)
        except ValueError:
            return False

        if len(ssn) != 9:
            return False
        for n in ssn:
            if n not in "0123456789":
                return False
        return True

class MissingInputSpecification(Exception):
    def __init__(self, input_name, message_fmt="Missing input specification for {input_name}"):
        self.input_name = input_name
        self.message = message_fmt.format(input_name=input_name)
        super().__init__(self.message)

class MissingInput(Exception):
    def __init__(self, input_name, message_fmt="Unsupplied Input: {input_name}"):
        self.input_name = input_name
        self.message = message_fmt.format(input_name=input_name)
        super().__init__(self.message)

class InvalidInput(Exception):
    def __init__(self, input_name, invalid_value, message_fmt="Invalid Input: {input_name} (currently: '{value}')"):
        self.input_name = input_name
        self.value = invalid_value
        self.message = message_fmt.format(input_name=input_name, value=invalid_value)
        super().__init__(self.message)

class InputStore(MutableMapping):
    def __init__(self, input_config, input_specs={}):
        if isinstance(input_config, configparser.ConfigParser):
            self.config = input_config
        else:
            self.config = configparser.ConfigParser()
            with open(input_config) as config_file:
                self.config.read_file(config_file)
        self.input_specs = input_specs

    def write(self, filename):
        with open(filename, 'w') as outfile:
            self.config.write(outfile)

    def update_input_spec(self, input_specs):
        self.input_specs = input_specs

    def provides(self, input_obj):
        return self.config.has_option(input_obj.section(), input_obj.base_name())

    def __getitem__(self, key):
        if key not in self.input_specs:
            raise MissingInputSpecification(key)

        i = self.input_specs[key]
        if not self.provides(i):
            raise MissingInput(key)
        string = self.config.get(i.section(), i.base_name())
        if not i.valid(string):
            raise InvalidInput(key, string)
        return i.value(string)

    def __setitem__(self, key, value):
        if key not in self.input_specs:
            raise MissingInputSpecification(key)
        i = self.input_specs[key]
        if i.section() not in self.config.sections():
            self.config.add_section(i.section())
        self.config.set(i.section(), i.base_name(), value)

    def __contains__(self, key):
        section, base_name = key.split(".")
        return self.config.has_option(section, base_name)

    def __delitem__(self, key):
        section, base_name = key.split(".")
        self.config.remove_option(section, base_name)
        if len(self.config[section]) == 0:
            self.config.remove_section(section)

    def __iter__(self):
        return iter(self.config)

    def __len__(self):
        return len(self.config)



from habutax.forms import ty2021
from habutax.forms import ty2022
from habutax.forms import ty2023

available_forms = {
    2021: ty2021.available_forms,
    2022: ty2022.available_forms,
    2023: ty2023.available_forms,
}



from enum import Enum


class StringyEnum(object):
    """
    Mixin which changes the string formatting to only include the name
    """

    def __str__(self):
        return self.name


def make(name, options):
    assert type(options) == dict
    for k, v in options.items():
        assert isinstance(k, str)
        assert isinstance(v, str)
    return Enum(name, options, type=StringyEnum)


filing_status_2021 = make('1040 Filing Status', {
    'Single': "single, unmarried, or legally separated",
    'MarriedFilingJointly': "married and filing a joint return",
    'MarriedFilingSeparately': "married and file a separate return",
    'HeadOfHousehold': "unmarried and provide a home for certain other person",
    'QualifyingWidowWidower': "generally filed if your spouse died in the two years previous to the tax year of this return, you didn't remarry before the end of this tax year, and you have a child or stepchild whom you can claim as a dependent (see Form 1040 instructions for more)"
})

filing_status = make('1040 Filing Status', {
    'Single': "single, unmarried, or legally separated",
    'MarriedFilingJointly': "married and filing a joint return",
    'MarriedFilingSeparately': "married and file a separate return",
    'HeadOfHousehold': "unmarried and provide a home for certain other person",
    'QualifyingSurvivingSpouse': "generally filed if your spouse died in the two years previous to the tax year of this return, you didn't remarry before the end of this tax year, and you have a child or stepchild whom you can claim as a dependent (see Form 1040 instructions for more)"
})

us_states = make('US States', {
    'AL': 'Alabama',
    'AK': 'Alaska',
    'AZ': 'Arizona',
    'AR': 'Arkansas',
    'CA': 'California',
    'CO': 'Colorado',
    'CT': 'Connecticut',
    'DE': 'Delaware',
    'FL': 'Florida',
    'GA': 'Georgia',
    'HI': 'Hawaii',
    'ID': 'Idaho',
    'IL': 'Illinois',
    'IN': 'Indiana',
    'IA': 'Iowa',
    'KS': 'Kansas',
    'KY': 'Kentucky',
    'LA': 'Louisiana',
    'ME': 'Maine',
    'MD': 'Maryland',
    'MA': 'Massachusetts',
    'MI': 'Michigan',
    'MN': 'Minnesota',
    'MS': 'Mississippi',
    'MO': 'Missouri',
    'MT': 'Montana',
    'NE': 'Nebraska',
    'NV': 'Nevada',
    'NH': 'New Hampshire',
    'NJ': 'New Jersey',
    'NM': 'New Mexico',
    'NY': 'New York',
    'NC': 'North Carolina',
    'ND': 'North Dakota',
    'OH': 'Ohio',
    'OK': 'Oklahoma',
    'OR': 'Oregon',
    'PA': 'Pennsylvania',
    'RI': 'Rhode Island',
    'SC': 'South Carolina',
    'SD': 'South Dakota',
    'TN': 'Tennessee',
    'TX': 'Texas',
    'UT': 'Utah',
    'VT': 'Vermont',
    'VA': 'Virginia',
    'WA': 'Washington',
    'WV': 'West Virginia',
    'WI': 'Wisconsin',
    'WY': 'Wyoming'
})

taxpayer_or_spouse = make('Taxpayer or Spouse', {
    'taxpayer': 'This belongs to the taxpayer',
    'spouse': 'This belongs to the taxpayer\'s spouse (if filing a joint return)'
})

taxpayer_spouse_or_both = make('Taxpayer, Spouse, or Both', {
    'taxpayer': 'This belongs to the taxpayer',
    'spouse': 'This belongs to the taxpayer\'s spouse (if filing a joint return)',
    'both': 'This belongs to both the taxpayer and their spouse (if filing a joint return)'
})



from habutax import fields
from habutax import form
from habutax import inputs
from habutax import values

class DependencyTracker(object):
    def __init__(self):
        # map names of unmet dependencies to a list of their outstanding
        # dependents
        self._unmet = {}

        # A list of the names of newly-satisfied dependencies
        self._met = []

    def add_unmet(self, dependency_name, dependent):
        """Add an unmet dependency for a particular dependent"""
        if dependency_name not in self._unmet:
            self._unmet[dependency_name] = [dependent]
        else:
            self._unmet[dependency_name].append(dependent)

    def has_met(self):
        """Return True if there are met dependencies which have not been
        removed by calling met_dependents()""" 
        return len(self._met) > 0

    def has_unmet(self):
        """Returns True if there are still unmet dependencies (does not count
        met dependencies which have not yet been removed via a call to
        met_dependents()"""
        for dependency, dependents in self._unmet.items():
            if len(dependents) > 0 and dependency not in self._met:
                return True
        return False

    def meet(self, dependency_name):
        """Mark a depdendency as having been satisfied (allows any dependents
        to be released in met_dependents()"""
        self._met.append(dependency_name)

    def unmet_dependencies(self):
        """Returns the names of the unmet dependencies"""
        return list(self._unmet.keys())

    def unmet_dependents(self, dependency):
        """Return the list of objects of the dependents waiting on an unmet dependency"""
        return self._unmet[dependency]

    def met_dependents(self):
        """Generator returning the objects representing all dependents whose
        dependencies have been marked as satisfied. Note that this also removes
        the returned dependency, so if a further dependency is later
        discovered, this object must be re-added using add_unmet()"""
        while len(self._met) > 0:
            met = self._met[0]
            # check if there are any dependents waiting on this satisfied dependency
            if met in self._unmet:
                dependent = self._unmet[met].pop()

                if len(self._unmet[met]) == 0:
                    del self._unmet[met]
                    self._met.pop(0)

                yield dependent
            else:
                self._met.pop(0)
                continue

def _sort_keys(key):
    last_numeric = False
    last_alpha = False
    current = ''
    keys = []

    for c in key:
        this_numeric = c in '01234567890'
        this_alpha = c.isalpha()
        if (last_numeric and not this_numeric) or \
                (last_alpha and not this_alpha):
            if len(current) > 0:
                if last_numeric:
                    keys.append((False, int(current)))
                else:
                    keys.append((True, current))
                current = ''

        if this_numeric or this_alpha:
            current += c
        else:
            current = ''

        last_numeric = this_numeric
        last_alpha = this_alpha

    # Output the last partial sub-key in the string
    if len(current) > 0:
        if last_numeric:
            keys.append((False, int(current)))
        else:
            keys.append((True, current))

    return keys

def sort_keys(key):
    if isinstance(key, fields.Field) or isinstance(key, inputs.Input):
        key = key.name()

    if '.' in key:
        form, key = key.split('.')
    else:
        form = ''
    return (_sort_keys(form), _sort_keys(key))

class Solver(object):
    def __init__(self, input_config, form_list, prompt=None):
        self._prompt = prompt
        self._refused_input = self._prompt is None

        # Create a map to easily look up the available forms by name
        self._form_map = {f.form_name: f for f in form_list}

        # Map all available input and field names to their objects
        self._input_map = {}
        self._field_map = {}

        # Input values (also modified with any values the user specifies
        # interactively)
        self._i = input_config

        # Current state of solver, including form instances, values calculated,
        # any unmet field/input dependencies
        self.forms = {}
        self._v = values.ValueStore()
        self._unattempted_fields = []
        self._unimplemented_fields = []
        self._solving_fields = set()
        self._field_dependencies = DependencyTracker()
        self._input_dependencies = DependencyTracker()

        self._done_solving = False # Set to True if/when done solving
        self._solved = False       # Set to True if/when successfully solved

    def _add_unattempted(self, unattempted):
        if isinstance(unattempted, list):
            self._unattempted_fields.extend(unattempted)
        else:
            self._unattempted_fields.append(unattempted)
        self._unattempted_fields.sort(key=sort_keys)

    def _add_form(self, form_name, input_only=False):
        """
        Add a form to those that the solver is aware of. This typically makes
        its fields and inputs available to to be used by other forms, and
        causes all its required fields to be attempted. However, if
        `input_only` is True, this does not cause the form to be filled out or
        any of its fields to be attempted (unless later used by other forms,
        meaning _add_form() will be called again. This `input_only` mode only
        causes inputs to be added, and is intended to allow for de-duplicating
        input felds when multiple forms need the same input.
        """
        form_name, form_instance = form.name_and_instance(form_name)

        if form_name not in self._form_map:
            raise NotImplementedError(f'Form {form_name} is not supported.')

        new_form = self._form_map[form_name](solver=self, instance=form_instance)

        # Add new inputs to our internal map of names to input objects, update
        # the input mapper so it understands how to read these inputs
        for i in new_form.inputs():
            assert i not in self._input_map
            self._input_map[i.name()] = i
        self._i.update_input_spec(self._input_map)

        # Don't add the form to our map of forms, any fields to our field map,
        # or any fields to the lists of unattempted fields, or fields we are
        # solving
        if input_only:
            return

        self.forms[new_form.name()] = new_form

        for f in new_form.fields():
            assert f not in self._field_map
            self._field_map[f.name()] = f

        self._add_unattempted(new_form.required_fields())
        self._solving_fields |= set([f.name() for f in new_form.required_fields()])

    def _add_input_spec(self, input_name):
        form_name, _ = input_name.split('.')
        self._add_form(form_name, input_only=True)

    def _attempt_input(self, input_name, needed_by):
        assert input_name in self._input_map
        missing = self._input_map[input_name]

        value, supplied = self._prompt(missing, needed_by)

        if supplied:
            assert missing.valid(value)
            self._i[missing.name()] = value
            self._input_dependencies.meet(missing.name())
        else:
            self._refused_input = True
        return supplied

    def _attempt_field(self, field):
        try:
            form_inputs = form.FormAccessor(self._i, field.form())
            form_values = form.FormAccessor(self._v, field.form())
            self._v[field.name()] = field.value(form_inputs, form_values)
            self._field_dependencies.meet(field.name())
        except values.UnmetDependency as ud:
            # If this field is not already in the fields the solver is
            # attempting to solve, add it
            if ud.dependency not in self._solving_fields:
                # If this field is not even in the list of forms the solver is
                # aware of, it must be in another form - find that form
                if ud.dependency not in self._field_map:
                    form_name, field_name = ud.dependency.split('.')
                    self._add_form(form_name)
                assert ud.dependency in self._field_map
                self._add_unattempted(self._field_map[ud.dependency])
                self._solving_fields.add(ud.dependency)

            self._field_dependencies.add_unmet(ud.dependency, field)
        except inputs.MissingInput as mi:
            self._input_dependencies.add_unmet(mi.input_name, field)
        except inputs.MissingInputSpecification as mis:
            self._add_input_spec(mis.input_name)
            return self._attempt_field(field)
        except fields.FieldNotImplemented as fni:
            self._unimplemented_fields.append(fni.field_name)

    def solve(self, form_names, field_names=[]):
        for form_name in form_names:
            self._add_form(form_name)

        # Add any specifically-requested fields to the list of fields we need
        # to attempt. This is probably primarily useful for testing (where we
        # may want to test a form which has primarily optional fields without
        # also including the forms which would typically include fields when
        # needed).
        for field_name in field_names:
            self._add_unattempted(self._field_map[field_name])
        self._solving_fields |= set(field_names)

        while len(self._unattempted_fields) > 0 \
                or self._input_dependencies.has_met() \
                or (self._input_dependencies.has_unmet() and not self._refused_input) \
                or self._field_dependencies.has_met():

            while len(self._unattempted_fields) > 0:
                self._attempt_field(self._unattempted_fields.pop())
            for field in sorted(self._field_dependencies.met_dependents(), key=sort_keys):
                self._attempt_field(field)
            if not self._refused_input:
                for input_name in sorted(self._input_dependencies.unmet_dependencies(), key=sort_keys):
                    needed_by = self._input_dependencies.unmet_dependents(input_name)
                    self._attempt_input(input_name, needed_by)
                    if self._refused_input:
                        break
            # Must be list() because _attempt_field modifies
            # input_dependencies, and we don't want to get stuck in a loop
            # waiting for an input we can't provide without polling user for it
            for field in list(self._input_dependencies.met_dependents()):
                self._attempt_field(field)

        assert len(self._unattempted_fields) == 0
        assert not self._input_dependencies.has_met()
        assert not self._field_dependencies.has_met()

        self._done_solving = True

        if not self._field_dependencies.has_unmet() \
                and not self._input_dependencies.has_unmet() \
                and len(self._unimplemented_fields) == 0:
            self._solved = True

        return self._solved

    def solution(self):
        """Return a ConfigParser object representing the portions of the
        requested forms which were successfully solved"""
        assert self._done_solving
        return self._v.to_config(self._field_map)

    def unimplemented_fields(self):
        """Return a list of the unimplemented field names after a call to
        solve()"""
        assert self._done_solving
        return self._unimplemented_fields

    def _unmet_dependencies(self, dependency_tracker):
        unmet_dependencies = {}
        for dep in dependency_tracker.unmet_dependencies():
            dependents = [f.name() for f in dependency_tracker.unmet_dependents(dep)]
            unmet_dependencies[dep] = dependents
        return unmet_dependencies

    def unmet_input_dependencies(self):
        """Return a dict mapping names of any unmet input dependencies to a
        list of the names of their (discovered) dependent fields after a call
        to solve()"""
        assert self._done_solving
        return self._unmet_dependencies(self._input_dependencies)

    def unmet_field_dependencies(self):
        """Return a dict mapping names of any unmet field dependencies to a
        list of the names of their (discovered) dependents after a call to
        solve()"""
        assert self._done_solving
        return self._unmet_dependencies(self._field_dependencies)



import habutax.enum as enum
from habutax.form import InputForm, Jurisdiction
from habutax.inputs import *

class Form1099R(InputForm):
    form_name = "1099-r"
    tax_year = 2023
    description = "Form 1099-R"
    long_description = "Distributions From Pensions, Annuities, Retirement or Profit-Sharing Plans, IRAs, Insurance Contracts, etc."
    jurisdiction = Jurisdiction.US

    def __init__(self, **kwargs):
        inputs = [
            StringInput('payer', description="Who paid you this distribution?"),
            StringInput('recipient', description="To whom was this distribution paid? (you or your spouse's name)"),
            EnumInput('belongs_to', enum.taxpayer_or_spouse, description="To whom was this distribution paid?"),
            FloatInput('box_1', description="Gross distribution"),
            FloatInput('box_2a', description="Taxable amount"),
            BooleanInput('box_2b_taxable_not_determined', description="'Taxable amount not determined' checked"),
            BooleanInput('box_2b_total_distribution', description="'Total distribution' checked"),
            FloatInput('box_3', description="Capital gain (included in box 2a)"),
            FloatInput('box_4', description="Federal income tax withheld"),
            FloatInput('box_5', description="Employee contributions/Designated Roth contributions or insurance premiums"),
            FloatInput('box_6', description="Net unrealized appreciation in employer’s securities"),
            StringInput('box_7_distirbution_codes', description="Distribution code(s)"),
            BooleanInput('box_7_ira_sep_simple', description="'IRA/SEP/SIMPLE' checked"),
            FloatInput('box_8', description="Other ($)"),
            FloatInput('box_8_pct', description="Other (%)"),
            FloatInput('box_9a', description="Your percentage of total distibution"),
            FloatInput('box_9b', description="Total employee contributions"),
            FloatInput('box_10', description="Amount allocable to IRR within 5 years"),
            FloatInput('box_11', description="1st year of desig. Roth contrib."),
            BooleanInput('box_12', description="'FATCA filing requirement' checked"),
            StringInput('box_13', description="Date of payment"),
            FloatInput('box_14_1', description="State tax withheld"),
            EnumInput('box_14_1_state', enum.us_states, allow_empty=True, description="State for which tax was withheld"),
            StringInput('box_15_1', description="State/Payer's state no."),
            FloatInput('box_16_1', description="State distribution"),
            FloatInput('box_14_2', description="State tax withheld"),
            EnumInput('box_14_2_state', enum.us_states, allow_empty=True, description="State for which tax was withheld"),
            StringInput('box_15_2', description="State/Payer's state no."),
            FloatInput('box_16_2', description="State distribution"),
            FloatInput('box_17_1', description="Local tax withheld"),
            StringInput('box_18_1', description="Name of locality"),
            FloatInput('box_19_1', description="Local distribution"),
            FloatInput('box_17_2', description="Local tax withheld"),
            StringInput('box_18_2', description="Name of locality"),
            FloatInput('box_19_2', description="Local distribution"),
        ]

        super().__init__(__class__, inputs, **kwargs)



import unittest

from habutax import enum


class EnumTestCase(unittest.TestCase):
    def test_enum_make(self):
        e = enum.make('a', {'b': "yay!", 'c': "foobar"})
        self.assertEqual(str(e.b), "b")
        self.assertEqual(str(e.c), "c")
        self.assertEqual(str(e['c']), "c")
        self.assertEqual(e.b.value, "yay!")
        self.assertEqual(e.c.name, "c")
        self.assertEqual(e.__members__['c'].value, "foobar")



import os

from habutax.enum import filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *
from habutax.pdf_fields import *

class FormNCD400SA(Form):
    form_name = "nc_d-400_sa"
    tax_year = 2021
    description = "D-400 Schedule A"
    long_description = "N.C. Itemized Deductions"
    jurisdiction = Jurisdiction.NC
    sequence_no = 2

    def __init__(self, **kwargs):
        inputs = [
            FloatInput('claim_of_right_income', description="Enter any 'Repayment of Claim of Right Income' which you can claim a NC deduction for (this is not common). See the instructions for NC D-400 Schedule A, line 8 for more information"),
        ]

        def nc_standard_deduction(self, i, v):
            if i['1040.filing_status'] in [filing_status.MarriedFilingJointly, filing_status.QualifyingWidowWidower]:
                return 21500.0
            elif i['1040.filing_status'] == filing_status.HeadOfHousehold:
                return 16125.0
            elif i['1040.filing_status'] == filing_status.Single:
                return 10750.0
            elif i['1040.filing_status'] == filing_status.MarriedFilingSeparately and not i['1040.standard_deduction_exceptions']:
                return 10750.0
            self.not_implemented()

        def line_1(self, i, v):
            mortgage_interest_points = sum([v[f'1098:{n}.box_1'] for n in range(i['1040.number_1098'])])
            mortgage_interest_points += sum([v[f'1098:{n}.box_6'] for n in range(i['1040.number_1098'])])
            return mortgage_interest_points

        def real_estate_tax_limit(self, i, v):
            return 5000.0 if i['1040.filing_status'] == filing_status.MarriedFilingSeparately else 10000.0

        optional_fields = [
            FloatField('nc_standard_deduction', nc_standard_deduction, places=0),
            FloatField('1', line_1, places=0),
            FloatField('2', lambda s, i, v: min(i['1040_sa.state_local_real_estate_taxes'], real_estate_tax_limit(s, i, v)), places=0),
            FloatField('3', lambda s, i, v: v['1'] + v['2'], places=0),
            FloatField('4', lambda s, i, v: 20000.0, places=0),
            FloatField('5', lambda s, i, v: min(v['3'], v['4']), places=0),
            FloatField('6', lambda s, i, v: min(i['1040_sa.charitable_cash_check'] + i['1040_sa.charitable_other_than_cash_check'], v['1040.11'] * 0.6), places=0),
            FloatField('7a', lambda s, i, v: i['1040_sa.medical_dental_expenses'], places=0),
            FloatField('7b', lambda s, i, v: v['nc_d-400.6'], places=0),
            FloatField('7c', lambda s, i, v: max(0.0, v['7b'] * 0.075), places=0),
            FloatField('7d', lambda s, i, v: 0.0 if v['7c'] > v['7a'] else v['7a'] - v['7c'], places=0),
            FloatField('8', lambda s, i, v: i['claim_of_right_income'], places=0),
            FloatField('9', lambda s, i, v: 0.0, places=0), # "reserved for future use"
            FloatField('10', lambda s, i, v: v['5'] + v['6'] + v['7d'] + v['8'] + v['9'], places=0),
        ]

        required_fields = [
            StringField('last_name', lambda s, i, v: i['1040.last_name'][:10]),
            StringField('ssn', lambda s, i, v: i['1040.you_ssn'][:3] + "-" + i['1040.you_ssn'][3:5] + "-" + i['1040.you_ssn'][5:]),
            FloatField('deduction', lambda s, i, v: max(v['nc_standard_deduction'], v['10']), places=0),
        ]

        pdf_fields = [
            TextPDFField('y_d400_sch_a_wf_lname', 'last_name', max_length=10),
            TextPDFField('y_d400_sch_a_wf_ssn', 'ssn'),
            TextPDFField('y_d400_sch_a_wf_li1', '1', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li2', '2', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li3', '3', max_length=8),
#            TextPDFField('y_d400_sch_a_wf_li4', '20000'),
            TextPDFField('y_d400_sch_a_wf_li5', '5', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li6', '6', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li7a', '7a', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li7b', '7b', max_length=9),
            TextPDFField('y_d400_sch_a_wf_li7c', '7c', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li7d', '7d', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li8', '8', max_length=8),
#            TextPDFField('y_d400_sch_a_wf_reserved', 'reserved_for_future_use', max_length=30),
#            TextPDFField('y_d400_sch_a_wf_li9', '9', max_length=8),
            TextPDFField('y_d400_sch_a_wf_li10', '10', max_length=8),
            #OptionlessButtonPDFField('z_d400schswf_clear', 'unknown'),
            #OptionlessButtonPDFField('z_d400_sch_a_wf_print', 'unknown'),
        ]
        pdf_file = os.path.join(os.path.dirname(__file__), 'fnc_d-400_sa.pdf')

        super().__init__(__class__, inputs, required_fields, optional_fields, pdf_fields=pdf_fields, pdf_file=pdf_file, **kwargs)

    def needs_filing(self, values):
        if 'nc_d-400_sa.deduction' not in values:
            return False
        return values['nc_d-400_sa.10'] > values['nc_d-400_sa.nc_standard_deduction']



import os

from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *

class FormNCD400ConsumerUseTaxWkst(Form):
    form_name = "nc_d-400_consumer_use_tax_wkst"
    tax_year = 2022
    description = "Consumer Use Tax Worksheet (NC Form D-400)"
    long_description = "Line 18"
    jurisdiction = Jurisdiction.NC

    def __init__(self, **kwargs):
        inputs = [
            BooleanInput('full_records', description="Do you have records of out-of-state (NC) purchases your or your spouse made for tax year 2022?"),
            FloatInput('out_of_state_purchases_pre_oct', description="Enter the total amount of out-of-state purchases, including delivery charges, for 1/1/2022 through 9/30/2022"),
            FloatInput('county_tax_pct_pre_oct', description="Enter '0.075' for Durham and Orange Counties; '0.0725' for Mecklenburg or Wake Counties; '0.07' for Alexander, Anson, Ashe, Bertie, Buncombe, Cabarrus, Catawba, Chatham, Cherokee, Clay, Cumberland, Davidson, Duplin, Edgecombe, Forsyth, Gaston, Graham, Greene, Halifax, Harnett, Haywood, Hertford, Jackson, Jones, Lee, Lincoln, Madison, Martin, Montgomery, Moore, New Hanover, Onslow, Pasquotank, Pitt, Randolph, Robeson, Rockingham, Rowan, Rutherford, Sampson, Stanly, Surry, Swain, and Wilkes Counties; and '0.0675' for all other North Carolina counties"),
            FloatInput('out_of_state_purchases_post_oct', description="Enter the total amount of out-of-state purchases, including delivery charges, for 10/1/2022 through 12/31/2022"),
            FloatInput('county_tax_pct_post_oct', description="Enter '0.075' for Durham and Orange Counties; '0.0725' for Mecklenburg or Wake Counties; '0.07' for Alexander, Alleghany, Anson, Ashe, Bertie, Buncombe, Cabarrus, Catawba, Chatham, Cherokee, Clay, Cumberland, Davidson, Duplin, Edgecombe, Forsyth, Gaston, Graham, Greene, Halifax, Harnett, Haywood, Hertford, Jackson, Jones, Lee, Lincoln, Madison, Martin, Montgomery, Moore, New Hanover, Onslow, Pasquotank, Pitt, Randolph, Robeson, Rockingham, Rowan, Rutherford, Sampson, Stanly, Surry, Swain, and Wilkes Counties; and '0.0675' for all other North Carolina counties"),
            # TODO post-10/1/2022 out-of-state-purchases
            FloatInput('other_state_sales_tax', description="Enter the amount of sales tax legally and properly paid to another state or North Carolina on your out-of-state purchases."),
        ]

        def estimate(self, i, v):
            use_tax_table = [
                (2200, 1),
                (3700, 2),
                (5200, 3),
                (6700, 4),
                (8100, 5),
                (9600, 6),
                (11100, 7),
                (12600, 8),
                (14100, 9),
                (15600, 10),
                (17000, 11),
                (18500, 12),
                (20000, 13),
                (21500, 14),
                (23000, 15),
                (24400, 16),
                (25900, 17),
                (27400, 18),
                (28900, 19),
                (30400, 20),
                (31900, 21),
                (33300, 22),
                (34800, 23),
                (36300, 24),
                (37800, 25),
                (39300, 26),
                (40700, 27),
                (42200, 28),
                (43700, 29),
                (45200, 30),
            ]
            for max_taxable_income, use_tax in use_tax_table:
                if v['nc_d-400.14'] < max_taxable_income:
                    return float(use_tax)
            assert v['nc_d-400.14'] >= 45200
            return v['nc_d-400.14'] * 0.000675

        optional_fields = [
            FloatField('estimate', estimate, places=0),
            FloatField('1', lambda s, i, v: i['out_of_state_purchases_pre_oct'], places=0),
            FloatField('2', lambda s, i, v: i['county_tax_pct_pre_oct'] * v['1'], places=0),
            FloatField('3', lambda s, i, v: i['out_of_state_purchases_post_oct'], places=0),
            FloatField('4', lambda s, i, v: i['county_tax_pct_post_oct'] * v['3'], places=0),
            FloatField('5', lambda s, i, v: min(i['other_state_sales_tax'], v['2'] + v['4']), places=0),
            FloatField('6', lambda s, i, v: round(v['2'] + v['4'] - v['5'], 0), places=0),
            FloatField('consumer_use_tax', lambda s, i, v: v['6'] if i['full_records'] else v['estimate'], places=0),
        ]

        required_fields = [
        ]

        super().__init__(__class__, inputs, required_fields, optional_fields, **kwargs)

    def needs_filing(self, values):
        return False



from habutax.enum import filing_status
from habutax.form import Form, Jurisdiction
from habutax.inputs import *
from habutax.fields import *

class FormNCD400ChildDeductionWkst(Form):
    form_name = "nc_d-400_child_deduction_wkst"
    tax_year = 2023
    description = "Child Deduction Worksheet (NC Form D-400)"
    long_description = "Line 10b"
    jurisdiction = Jurisdiction.NC

    def __init__(self, **kwargs):
        inputs = [
        ]

        def deduction_per_child(self, i, v):
            status = i['1040.filing_status']
            federal_agi = v['2']

            if status in [filing_status.MarriedFilingJointly, filing_status.QualifyingSurvivingSpouse]:
                if federal_agi <= 40000:
                    return 3000.0
                elif federal_agi <= 60000:
                    return 2500.0
                elif federal_agi <= 80000:
                    return 2000.0
                elif federal_agi <= 100000:
                    return 1500.0
                elif federal_agi <= 120000:
                    return 1000.0
                elif federal_agi <= 140000:
                    return 500.0
                else:
                    return 0.0
            elif status == filing_status.HeadOfHousehold:
                if federal_agi <= 30000:
                    return 3000.0
                elif federal_agi <= 45000:
                    return 2500.0
                elif federal_agi <= 60000:
                    return 2000.0
                elif federal_agi <= 75000:
                    return 1500.0
                elif federal_agi <= 90000:
                    return 1000.0
                elif federal_agi <= 105000:
                    return 500.0
                else:
                    return 0.0
            elif status in [filing_status.Single, filing_status.MarriedFilingSeparately]:
                if federal_agi <= 20000:
                    return 3000.0
                elif federal_agi <= 30000:
                    return 2500.0
                elif federal_agi <= 40000:
                    return 2000.0
                elif federal_agi <= 50000:
                    return 1500.0
                elif federal_agi <= 60000:
                    return 1000.0
                elif federal_agi <= 70000:
                    return 500.0
                else:
                    return 0.0
            else:
                self.not_implemented()

        optional_fields = [
            StringField('1', lambda s, i, v: str(i['1040.filing_status'])),
            FloatField('2', lambda s, i, v: v['nc_d-400.6'], places=0),
            IntegerField('3', lambda s, i, v: sum([v[f'1040.dependent_{n}_ctc'] for n in range(i['1040.number_dependents'])])),
            FloatField('4', deduction_per_child, places=0),
            FloatField('5', lambda s, i, v: v['3'] * v['4'], places=0),
        ]

        required_fields = [
        ]

        super().__init__(__class__, inputs, required_fields, optional_fields,
                         **kwargs)

    def needs_filing(self, values):
        return False
